"""
String functions.

.. versionadded:: 1.1.0
"""

import html
import math
import re
import typing
import typing as t
import unicodedata
from urllib.parse import parse_qsl, urlencode, urlsplit, urlunsplit

import pydash as pyd

from .helpers import UNSET, Unset
from .types import NumberT


__all__ = (
    "camel_case",
    "capitalize",
    "chop",
    "chop_right",
    "chars",
    "clean",
    "count_substr",
    "deburr",
    "decapitalize",
    "ends_with",
    "ensure_ends_with",
    "ensure_starts_with",
    "escape",
    "escape_reg_exp",
    "has_substr",
    "human_case",
    "insert_substr",
    "join",
    "kebab_case",
    "lines",
    "lower_case",
    "lower_first",
    "number_format",
    "pad",
    "pad_end",
    "pad_start",
    "pascal_case",
    "predecessor",
    "prune",
    "quote",
    "reg_exp_js_match",
    "reg_exp_js_replace",
    "reg_exp_replace",
    "repeat",
    "replace",
    "replace_end",
    "replace_start",
    "separator_case",
    "series_phrase",
    "series_phrase_serial",
    "slugify",
    "snake_case",
    "split",
    "start_case",
    "starts_with",
    "strip_tags",
    "substr_left",
    "substr_left_end",
    "substr_right",
    "substr_right_end",
    "successor",
    "surround",
    "swap_case",
    "title_case",
    "to_lower",
    "to_upper",
    "trim",
    "trim_end",
    "trim_start",
    "truncate",
    "unescape",
    "unquote",
    "upper_case",
    "upper_first",
    "url",
    "words",
)

T = t.TypeVar("T")
T2 = t.TypeVar("T2")


class JSRegExp:
    """
    Javascript-style regular expression pattern.

    Converts a Javascript-style regular expression to the equivalent Python version.
    """

    def __init__(self, reg_exp: str) -> None:
        pattern, options = reg_exp[1:].rsplit("/", 1)

        self._global = "g" in options
        self._ignore_case = "i" in options

        flags = re.I if self._ignore_case else 0
        self.pattern = re.compile(pattern, flags=flags)

    def find(self, text: str) -> t.List[str]:
        """Return list of regular expression matches."""
        if self._global:
            results = self.pattern.findall(text)
        else:
            res = self.pattern.search(text)
            if res:
                results = [res.group()]
            else:
                results = []
        return results

    def replace(self, text: str, repl: t.Union[str, t.Callable[[re.Match], str]]) -> str:
        """Replace parts of text that match the regular expression."""
        count = 0 if self._global else 1
        return self.pattern.sub(repl, text, count=count)


HTML_ESCAPES = {"&": "&amp;", "<": "&lt;", ">": "&gt;", '"': "&quot;", "'": "&#39;", "`": "&#96;"}

DEBURRED_LETTERS = {
    "\xC0": "A",
    "\xC1": "A",
    "\xC2": "A",
    "\xC3": "A",
    "\xC4": "A",
    "\xC5": "A",
    "\xE0": "a",
    "\xE1": "a",
    "\xE2": "a",
    "\xE3": "a",
    "\xE4": "a",
    "\xE5": "a",
    "\xC7": "C",
    "\xE7": "c",
    "\xD0": "D",
    "\xF0": "d",
    "\xC8": "E",
    "\xC9": "E",
    "\xCA": "E",
    "\xCB": "E",
    "\xE8": "e",
    "\xE9": "e",
    "\xEA": "e",
    "\xEB": "e",
    "\xCC": "I",
    "\xCD": "I",
    "\xCE": "I",
    "\xCF": "I",
    "\xEC": "i",
    "\xED": "i",
    "\xEE": "i",
    "\xEF": "i",
    "\xD1": "N",
    "\xF1": "n",
    "\xD2": "O",
    "\xD3": "O",
    "\xD4": "O",
    "\xD5": "O",
    "\xD6": "O",
    "\xD8": "O",
    "\xF2": "o",
    "\xF3": "o",
    "\xF4": "o",
    "\xF5": "o",
    "\xF6": "o",
    "\xF8": "o",
    "\xD9": "U",
    "\xDA": "U",
    "\xDB": "U",
    "\xDC": "U",
    "\xF9": "u",
    "\xFA": "u",
    "\xFB": "u",
    "\xFC": "u",
    "\xDD": "Y",
    "\xFD": "y",
    "\xFF": "y",
    "\xC6": "Ae",
    "\xE6": "ae",
    "\xDE": "Th",
    "\xFE": "th",
    "\xDF": "ss",
    "\xD7": " ",
    "\xF7": " ",
}

# Use Javascript style regex to make Lo-Dash compatibility easier.
# Lodash Regex definitions: https://github.com/lodash/lodash/blob/master/.internal/unicodeWords.js

# References: https://github.com/lodash/lodash/blob/master/words.js#L8
RS_ASCII_WORDS = "/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g"
RS_LATIN1 = "/[\xC0-\xFF]/g"

# Used to compose unicode character classes.
RS_ASTRAL_RANGE = "\\ud800-\\udfff"
RS_COMBO_MARKS_RANGE = "\\u0300-\\u036f"
RE_COMBO_HALF_MARKS_RANGE = "\\ufe20-\\ufe2f"
RS_COMBO_SYMBOLS_RANGE = "\\u20d0-\\u20ff"
RS_COMBO_MARKS_EXTENDED_RANGE = "\\u1ab0-\\u1aff"
RS_COMBO_MARKS_SUPPLEMENT_RANGE = "\\u1dc0-\\u1dff"
RS_COMBO_RANGE = (
    RS_COMBO_MARKS_RANGE
    + RE_COMBO_HALF_MARKS_RANGE
    + RS_COMBO_SYMBOLS_RANGE
    + RS_COMBO_MARKS_EXTENDED_RANGE
    + RS_COMBO_MARKS_SUPPLEMENT_RANGE
)
RS_DINGBAT_RANGE = "\\u2700-\\u27bf"
RS_LOWER_RANGE = "a-z\\xdf-\\xf6\\xf8-\\xff"
RS_MATH_OP_RANGE = "\\xac\\xb1\\xd7\\xf7"
RS_NON_CHAR_RANGE = "\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf"
RS_PUNCTUATION_RANGE = "\\u2000-\\u206f"
RS_SPACE_RANGE = (
    " \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\"
    "u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\"
    "u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000"
)
RS_UPPER_RANGE = "A-Z\\xc0-\\xd6\\xd8-\\xde"
RS_VAR_RANGE = "\\ufe0e\\ufe0f"
RS_BREAK_RANGE = RS_MATH_OP_RANGE + RS_NON_CHAR_RANGE + RS_PUNCTUATION_RANGE + RS_SPACE_RANGE

# Used to compose unicode capture groups.
RS_APOS = "['\u2019]"
RS_BREAK = f"[{RS_BREAK_RANGE}]"
RS_COMBO = f"[{RS_COMBO_RANGE}]"
RS_DIGIT = "\\d"
RS_DINGBAT = f"[{RS_DINGBAT_RANGE}]"
RS_LOWER = f"[{RS_LOWER_RANGE}]"
RS_MISC = (
    f"[^{RS_ASTRAL_RANGE}{RS_BREAK_RANGE}{RS_DIGIT}"
    f"{RS_DINGBAT_RANGE}{RS_LOWER_RANGE}{RS_UPPER_RANGE}]"
)
RS_FITZ = "\\ud83c[\\udffb-\\udfff]"
RS_MODIFIER = f"(?:{RS_COMBO}|{RS_FITZ})"
RS_NON_ASTRAL = f"[^{RS_ASTRAL_RANGE}]"
RS_REGIONAL = "(?:\\ud83c[\\udde6-\\uddff]){2}"
RS_SURR_PAIR = "[\\ud800-\\udbff][\\udc00-\\udfff]"
RS_UPPER = f"[{RS_UPPER_RANGE}]"
RS_ZWJ = "\\u200d"

# Used to compose unicode regexes.
RS_MISC_LOWER = f"(?:{RS_LOWER}|{RS_MISC})"
RS_MISC_UPPER = f"(?:{RS_UPPER}|{RS_MISC})"
RS_OPT_CONTR_LOWER = f"(?:{RS_APOS}(?:d|ll|m|re|s|t|ve))?"
RS_OPT_CONTR_UPPER = f"(?:{RS_APOS}(?:D|LL|M|RE|S|T|VE))?"
RE_OPT_MOD = f"{RS_MODIFIER}?"
RS_OPT_VAR = f"[{RS_VAR_RANGE}]?"
RS_OPT_JOIN = (
    f"(?:{RS_ZWJ}(?:{RS_NON_ASTRAL}|{RS_REGIONAL}|{RS_SURR_PAIR}){RS_OPT_VAR}{RE_OPT_MOD})*"
)
RS_ORD_LOWER = "\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])"
RS_ORD_UPPER = "\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])"
RS_SEQ = RS_OPT_VAR + RE_OPT_MOD + RS_OPT_JOIN
RS_EMOJI = f"(?:{RS_DINGBAT}|{RS_REGIONAL}|{RS_SURR_PAIR}){RS_SEQ}"

RS_HAS_UNICODE_WORD = "[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]"
RS_UNICODE_WORDS = (
    f"/"
    f"{RS_UPPER}?{RS_LOWER}+{RS_OPT_CONTR_LOWER}(?={RS_BREAK}|{RS_UPPER}|$)"
    f"|{RS_MISC_UPPER}+{RS_OPT_CONTR_UPPER}(?={RS_BREAK}|{RS_UPPER}{RS_MISC_LOWER}|$)"
    f"|{RS_UPPER}?{RS_MISC_LOWER}+{RS_OPT_CONTR_LOWER}"
    f"|{RS_UPPER}+{RS_OPT_CONTR_UPPER}"
    f"|{RS_ORD_UPPER}"
    f"|{RS_ORD_LOWER}"
    f"|{RS_DIGIT}+"
    f"|{RS_EMOJI}"
    f"/g"
)

# Compiled regexes for use in functions.
JS_RE_ASCII_WORDS = JSRegExp(RS_ASCII_WORDS)
JS_RE_UNICODE_WORDS = JSRegExp(RS_UNICODE_WORDS)
JS_RE_LATIN1 = JSRegExp(RS_LATIN1)
RE_HAS_UNICODE_WORD = re.compile(RS_HAS_UNICODE_WORD)
RE_APOS = re.compile(RS_APOS)
RE_HTML_TAGS = re.compile(r"<\/?[^>]+>")


def camel_case(text: t.Any) -> str:
    """
    Converts `text` to camel case.

    Args:
        text: String to convert.

    Returns:
        String converted to camel case.

    Example:

        >>> camel_case('FOO BAR_bAz')
        'fooBarBAz'

    .. versionadded:: 1.1.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    text = "".join(word.title() for word in compounder(text))
    return text[:1].lower() + text[1:]


def capitalize(text: t.Any, strict: bool = True) -> str:
    """
    Capitalizes the first character of `text`.

    Args:
        text: String to capitalize.
        strict: Whether to cast rest of string to lower case. Defaults to ``True``.

    Returns:
        Capitalized string.

    Example:

        >>> capitalize('once upon a TIME')
        'Once upon a time'
        >>> capitalize('once upon a TIME', False)
        'Once upon a TIME'

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Added `strict` option.
    """
    text = pyd.to_string(text)
    return text.capitalize() if strict else text[:1].upper() + text[1:]


def chars(text: t.Any) -> t.List[str]:
    """
    Split `text` into a list of single characters.

    Args:
        text: String to split up.

    Returns:
        List of individual characters.

    Example:

        >>> chars('onetwo')
        ['o', 'n', 'e', 't', 'w', 'o']

    .. versionadded:: 3.0.0
    """
    return list(pyd.to_string(text))


def chop(text: t.Any, step: int) -> t.List[str]:
    """
    Break up `text` into intervals of length `step`.

    Args:
        text: String to chop.
        step: Interval to chop `text`.

    Returns:
        List of chopped characters. If `text` is `None` an empty list is returned.

    Example:

        >>> chop('abcdefg', 3)
        ['abc', 'def', 'g']

    .. versionadded:: 3.0.0
    """
    if text is None:
        return []

    text = pyd.to_string(text)

    if step <= 0:
        chopped = [text]
    else:
        chopped = [text[i : i + step] for i in range(0, len(text), step)]

    return chopped


def chop_right(text: t.Any, step: int) -> t.List[str]:
    """
    Like :func:`chop` except `text` is chopped from right.

    Args:
        text: String to chop.
        step: Interval to chop `text`.

    Returns:
        List of chopped characters.

    Example:

        >>> chop_right('abcdefg', 3)
        ['a', 'bcd', 'efg']

    .. versionadded:: 3.0.0
    """
    if text is None:
        return []

    text = pyd.to_string(text)

    if step <= 0:
        chopped = [text]
    else:
        text_len = len(text)
        chopped = [text[-(i + step) : text_len - i] for i in range(0, text_len, step)][::-1]

    return chopped


def clean(text: t.Any) -> str:
    """
    Trim and replace multiple spaces with a single space.

    Args:
        text: String to clean.

    Returns:
        Cleaned string.

    Example:

        >>> clean('a  b   c    d')
        'a b c d'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return " ".join(pyd.compact(text.split()))


def count_substr(text: t.Any, subtext: t.Any) -> int:
    """
    Count the occurrences of `subtext` in `text`.

    Args:
        text: Source string to count from.
        subtext: String to count.

    Returns:
        Number of occurrences of `subtext` in `text`.

    Example:

        >>> count_substr('aabbccddaabbccdd', 'bc')
        2

    .. versionadded:: 3.0.0
    """
    if text is None or subtext is None:
        return 0

    text = pyd.to_string(text)
    subtext = pyd.to_string(subtext)

    return text.count(subtext)


def deburr(text: t.Any) -> str:
    """
    Deburrs `text` by converting latin-1 supplementary letters to basic latin letters.

    Args:
        text: String to deburr.

    Returns:
        Deburred string.

    Example:

        >>> deburr('déjà vu')
        '...
        >>> 'deja vu'
        'deja vu'

    .. versionadded:: 2.0.0
    """
    text = pyd.to_string(text)
    return JS_RE_LATIN1.replace(
        text, lambda match: DEBURRED_LETTERS.get(match.group(), match.group())
    )


def decapitalize(text: t.Any) -> str:
    """
    Decaptitalizes the first character of `text`.

    Args:
        text: String to decapitalize.

    Returns:
        Decapitalized string.

    Example:

        >>> decapitalize('FOO BAR')
        'fOO BAR'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text[:1].lower() + text[1:]


def ends_with(text: t.Any, target: t.Any, position: t.Union[int, None] = None) -> bool:
    """
    Checks if `text` ends with a given target string.

    Args:
        text: String to check.
        target: String to check for.
        position: Position to search from. Defaults to end of `text`.

    Returns:
        Whether `text` ends with `target`.

    Example:

        >>> ends_with('abc def', 'def')
        True
        >>> ends_with('abc def', 4)
        False

    .. versionadded:: 1.1.0
    """
    target = pyd.to_string(target)
    text = pyd.to_string(text)

    if position is None:
        position = len(text)

    return text[:position].endswith(target)


def ensure_ends_with(text: t.Any, suffix: t.Any) -> str:
    """
    Append a given suffix to a string, but only if the source string does not end with that suffix.

    Args:
        text: Source string to append `suffix` to.
        suffix: String to append to the source string if the source string does not end with
            `suffix`.

    Returns:
        source string possibly extended by `suffix`.

    Example:

        >>> ensure_ends_with('foo bar', '!')
        'foo bar!'
        >>> ensure_ends_with('foo bar!', '!')
        'foo bar!'

    .. versionadded:: 2.4.0
    """
    text = pyd.to_string(text)
    suffix = pyd.to_string(suffix)
    if text.endswith(suffix):
        return text
    return f"{text}{suffix}"


def ensure_starts_with(text: t.Any, prefix: t.Any) -> str:
    """
    Prepend a given prefix to a string, but only if the source string does not start with that
    prefix.

    Args:
        text: Source string to prepend `prefix` to.
        prefix: String to prepend to the source string if the source string does not start
            with `prefix`.

    Returns:
        source string possibly prefixed by `prefix`

    Example:

        >>> ensure_starts_with('foo bar', 'Oh my! ')
        'Oh my! foo bar'
        >>> ensure_starts_with('Oh my! foo bar', 'Oh my! ')
        'Oh my! foo bar'

    .. versionadded:: 2.4.0
    """
    text = pyd.to_string(text)
    prefix = pyd.to_string(prefix)
    if text.startswith(prefix):
        return text
    return f"{prefix}{text}"


def escape(text: t.Any) -> str:
    r"""
    Converts the characters ``&``, ``<``, ``>``, ``"``, ``'``, and ``\``` in `text` to their
    corresponding HTML entities.

    Args:
        text: String to escape.

    Returns:
        HTML escaped string.

    Example:

        >>> escape('"1 > 2 && 3 < 4"')
        '&quot;1 &gt; 2 &amp;&amp; 3 &lt; 4&quot;'

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Moved function to :mod:`pydash.strings`.
    """
    text = pyd.to_string(text)
    # NOTE: Not using html.escape because Lo-Dash escapes certain chars differently (e.g. `'` isn't
    # escaped by html.escape() but is by Lo-Dash).
    return "".join(HTML_ESCAPES.get(char, char) for char in text)


def escape_reg_exp(text: t.Any) -> str:
    """
    Escapes the RegExp special characters in `text`.

    Args:
        text: String to escape.

    Returns:
        RegExp escaped string.

    Example:

        >>> escape_reg_exp('[()]')
        '\\\\[\\\\(\\\\)\\\\]'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``escape_re``
    """
    text = pyd.to_string(text)
    return re.escape(text)


def has_substr(text: t.Any, subtext: t.Any) -> bool:
    """
    Returns whether `subtext` is included in `text`.

    Args:
        text: String to search.
        subtext: String to search for.

    Returns:
        Whether `subtext` is found in `text`.

    Example:

        >>> has_substr('abcdef', 'bc')
        True
        >>> has_substr('abcdef', 'bb')
        False

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    subtext = pyd.to_string(subtext)
    return text.find(subtext) >= 0


def human_case(text: t.Any) -> str:
    """
    Converts `text` to human case which has only the first letter capitalized and each word
    separated by a space.

    Args:
        text: String to convert.

    Returns:
        String converted to human case.

    Example:

        >>> human_case('abc-def_hij lmn')
        'Abc def hij lmn'
        >>> human_case('user_id')
        'User'

    .. versionadded:: 3.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return (
        pyd.chain(text)
        .snake_case()
        .reg_exp_replace("_id$", "")
        .replace("_", " ")
        .capitalize()
        .value()
    )


def insert_substr(text: t.Any, index: int, subtext: t.Any) -> str:
    """
    Insert `subtext` in `text` starting at position `index`.

    Args:
        text: String to add substring to.
        index: String index to insert into.
        subtext: String to insert.

    Returns:
        Modified string.

    Example:

        >>> insert_substr('abcdef', 3, '--')
        'abc--def'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    subtext = pyd.to_string(subtext)
    return text[:index] + subtext + text[index:]


def join(array: t.Iterable[t.Any], separator: t.Any = "") -> str:
    """
    Joins an iterable into a string using `separator` between each element.

    Args:
        array: Iterable to implode.
        separator: Separator to using when joining. Defaults to ``''``.

    Returns:
        Joined string.

    Example:

        >>> join(['a', 'b', 'c']) == 'abc'
        True
        >>> join([1, 2, 3, 4], '&') == '1&2&3&4'
        True
        >>> join('abcdef', '-') == 'a-b-c-d-e-f'
        True

    .. versionadded:: 2.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``implode``.
    """
    return pyd.to_string(separator).join(pyd.map_(array or (), pyd.to_string))


def kebab_case(text: t.Any) -> str:
    """
    Converts `text` to kebab case (a.k.a. spinal case).

    Args:
        text: String to convert.

    Returns:
        String converted to kebab case.

    Example:

        >>> kebab_case('a b c_d-e!f')
        'a-b-c-d-e-f'

    .. versionadded:: 1.1.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return "-".join(word.lower() for word in compounder(text) if word)


def lines(text: t.Any) -> t.List[str]:
    r"""
    Split lines in `text` into an array.

    Args:
        text: String to split.

    Returns:
        String split by lines.

    Example:

        >>> lines('a\nb\r\nc')
        ['a', 'b', 'c']

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.splitlines()


def lower_case(text: t.Any) -> str:
    """
    Converts string to lower case as space separated words.

    Args:
        text: String to convert.

    Returns:
        String converted to lower case as space separated words.

    Example:

        >>> lower_case('fooBar')
        'foo bar'
        >>> lower_case('--foo-Bar--')
        'foo bar'
        >>> lower_case('/?*Foo10/;"B*Ar')
        'foo 10 b ar'

    .. versionadded:: 4.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return " ".join(compounder(text)).lower()


def lower_first(text: str) -> str:
    """
    Converts the first character of string to lower case.

    Args:
        text: String passed in by the user.

    Returns:
        String in which the first character is converted to lower case.

    Example:

        >>> lower_first('FRED')
        'fRED'
        >>> lower_first('Foo Bar')
        'foo Bar'
        >>> lower_first('1foobar')
        '1foobar'
        >>> lower_first(';foobar')
        ';foobar'

    .. versionadded:: 4.0.0
    """
    return text[:1].lower() + text[1:]


def number_format(
    number: NumberT, scale: int = 0, decimal_separator: str = ".", order_separator: str = ","
) -> str:
    """
    Format a number to scale with custom decimal and order separators.

    Args:
        number: Number to format.
        scale: Number of decimals to include. Defaults to ``0``.
        decimal_separator: Decimal separator to use. Defaults to ``'.'``.
        order_separator: Order separator to use. Defaults to ``','``.

    Returns:
        Number formatted as string.

    Example:

        >>> number_format(1234.5678)
        '1,235'
        >>> number_format(1234.5678, 2, ',', '.')
        '1.234,57'

    .. versionadded:: 3.0.0
    """
    # Create a string formatter which converts number to the appropriately scaled representation.
    fmt = f"{{0:.{scale:d}f}}"

    try:
        num_parts = fmt.format(number).split(".")
    except ValueError:
        text = ""
    else:
        int_part = num_parts[0]
        dec_part = (num_parts + [""])[1]

        # Reverse the integer part, chop it into groups of 3, join on `order_separator`, and then
        # un-reverse the string.
        int_part = order_separator.join(chop(int_part[::-1], 3))[::-1]

        text = decimal_separator.join(pyd.compact([int_part, dec_part]))

    return text


def pad(text: t.Any, length: int, chars: t.Any = " ") -> str:
    """
    Pads `text` on the left and right sides if it is shorter than the given padding length. The
    `chars` string may be truncated if the number of padding characters can't be evenly divided by
    the padding length.

    Args:
        text: String to pad.
        length: Amount to pad.
        chars: Characters to pad with. Defaults to ``" "``.

    Returns:
        Padded string.

    Example:

        >>> pad('abc', 5)
        ' abc '
        >>> pad('abc', 6, 'x')
        'xabcxx'
        >>> pad('abc', 5, '...')
        '.abc.'

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Fix handling of multiple `chars` so that padded string isn't over padded.
    """
    # pylint: disable=redefined-outer-name
    text = pyd.to_string(text)
    text_len = len(text)

    if text_len >= length:
        return text

    mid = (length - text_len) / 2.0
    left_len = int(math.floor(mid))
    right_len = int(math.ceil(mid))
    chars = pad_end("", right_len, chars)

    return chars[:left_len] + text + chars


def pad_end(text: t.Any, length: int, chars: t.Any = " ") -> str:
    """
    Pads `text` on the right side if it is shorter than the given padding length. The `chars` string
    may be truncated if the number of padding characters can't be evenly divided by the padding
    length.

    Args:
        text: String to pad.
        length: Amount to pad.
        chars: Characters to pad with. Defaults to ``" "``.

    Returns:
        Padded string.

    Example:

        >>> pad_end('abc', 5)
        'abc  '
        >>> pad_end('abc', 5, '.')
        'abc..'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Renamed from ``pad_right`` to ``pad_end``.
    """
    # pylint: disable=redefined-outer-name
    text = pyd.to_string(text)
    length = max((length, len(text)))
    return (text + repeat(chars, length))[:length]


def pad_start(text: t.Any, length: int, chars: t.Any = " ") -> str:
    """
    Pads `text` on the left side if it is shorter than the given padding length. The `chars` string
    may be truncated if the number of padding characters can't be evenly divided by the padding
    length.

    Args:
        text: String to pad.
        length: Amount to pad.
        chars: Characters to pad with. Defaults to ``" "``.

    Returns:
        Padded string.

    Example:

        >>> pad_start('abc', 5)
        '  abc'
        >>> pad_start('abc', 5, '.')
        '..abc'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Renamed from ``pad_left`` to ``pad_start``.
    """
    # pylint: disable=redefined-outer-name
    text = pyd.to_string(text)
    length = max(length, len(text))
    return (repeat(chars, length) + text)[-length:]


def pascal_case(text: t.Any, strict: bool = True) -> str:
    """
    Like :func:`camel_case` except the first letter is capitalized.

    Args:
        text: String to convert.
        strict: Whether to cast rest of string to lower case. Defaults to ``True``.

    Returns:
        String converted to class case.

    Example:

        >>> pascal_case('FOO BAR_bAz')
        'FooBarBaz'
        >>> pascal_case('FOO BAR_bAz', False)
        'FooBarBAz'

    .. versionadded:: 3.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    text = pyd.to_string(text)

    if strict:
        text = text.lower()

    return capitalize(camel_case(text), strict=False)


def predecessor(char: t.Any) -> str:
    """
    Return the predecessor character of `char`.

    Args:
        char: Character to find the predecessor of.

    Returns:
        Predecessor character.

    Example:

        >>> predecessor('c')
        'b'
        >>> predecessor('C')
        'B'
        >>> predecessor('3')
        '2'

    .. versionadded:: 3.0.0
    """
    char = pyd.to_string(char)
    return chr(ord(char) - 1)


def prune(text: t.Any, length: int = 0, omission: str = "...") -> str:
    """
    Like :func:`truncate` except it ensures that the pruned string doesn't exceed the original
    length, i.e., it avoids half-chopped words when truncating. If the pruned text + `omission` text
    is longer than the original text, then the original text is returned.

    Args:
        text: String to prune.
        length: Target prune length. Defaults to ``0``.
        omission: Omission text to append to the end of the pruned string. Defaults
            to ``'...'``.

    Returns:
        Pruned string.

    Example:

        >>> prune('Fe fi fo fum', 5)
        'Fe fi...'
        >>> prune('Fe fi fo fum', 6)
        'Fe fi...'
        >>> prune('Fe fi fo fum', 7)
        'Fe fi...'
        >>> prune('Fe fi fo fum', 8, ',,,')
        'Fe fi fo,,,'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    text_len = len(text)
    omission_len = len(omission)

    if text_len <= length:
        return text

    # Replace non-alphanumeric chars with whitespace.
    def repl(match):
        char = match.group(0)
        return " " if char.upper() == char.lower() else char

    subtext = reg_exp_replace(text[: length + 1], r".(?=\W*\w*$)", repl)

    if re.match(r"\w\w", subtext[-2:]):
        # Last two characters are alphanumeric. Remove last "word" from end of string so that we
        # prune to the next whole word.
        subtext = reg_exp_replace(subtext, r"\s*\S+$", "")
    else:
        # Last character (at least) is whitespace. So remove that character as well as any other
        # whitespace.
        subtext = subtext[:-1].rstrip()

    subtext_len = len(subtext)

    # Only add omission text if doing so will result in a string that is equal to or smaller than
    # the original.
    if (subtext_len + omission_len) <= text_len:
        text = text[:subtext_len] + omission

    return text


def quote(text: t.Any, quote_char: t.Any = '"') -> str:
    """
    Quote a string with another string.

    Args:
        text: String to be quoted.
        quote_char: the quote character. Defaults to ``'"'``.

    Returns:
        the quoted string.

    Example:

        >>> quote('To be or not to be')
        '"To be or not to be"'
        >>> quote('To be or not to be', "'")
        "'To be or not to be'"

    .. versionadded:: 2.4.0
    """
    return surround(text, quote_char)


def reg_exp_js_match(text: t.Any, reg_exp: str) -> t.List[str]:
    """
    Return list of matches using Javascript style regular expression.

    Args:
        text: String to evaluate.
        reg_exp: Javascript style regular expression.

    Returns:
        List of matches.

    Example:

        >>> reg_exp_js_match('aaBBcc', '/bb/')
        []
        >>> reg_exp_js_match('aaBBcc', '/bb/i')
        ['BB']
        >>> reg_exp_js_match('aaBBccbb', '/bb/i')
        ['BB']
        >>> reg_exp_js_match('aaBBccbb', '/bb/gi')
        ['BB', 'bb']

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `text` first.

    .. versionchanged:: 4.0.0
        Renamed from ``js_match`` to ``reg_exp_js_match``.
    """
    text = pyd.to_string(text)
    return JSRegExp(reg_exp).find(text)


def reg_exp_js_replace(
    text: t.Any, reg_exp: str, repl: t.Union[str, t.Callable[[re.Match], str]]
) -> str:
    """
    Replace `text` with `repl` using Javascript style regular expression to find matches.

    Args:
        text: String to evaluate.
        reg_exp: Javascript style regular expression.
        repl: Replacement string or callable.

    Returns:
        Modified string.

    Example:

        >>> reg_exp_js_replace('aaBBcc', '/bb/', 'X')
        'aaBBcc'
        >>> reg_exp_js_replace('aaBBcc', '/bb/i', 'X')
        'aaXcc'
        >>> reg_exp_js_replace('aaBBccbb', '/bb/i', 'X')
        'aaXccbb'
        >>> reg_exp_js_replace('aaBBccbb', '/bb/gi', 'X')
        'aaXccX'

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `text` first.

    .. versionchanged:: 4.0.0
        Renamed from ``js_replace`` to ``reg_exp_js_replace``.
    """
    text = pyd.to_string(text)
    if not pyd.is_function(repl):
        repl = pyd.to_string(repl)
    return JSRegExp(reg_exp).replace(text, repl)


def reg_exp_replace(
    text: t.Any,
    pattern: t.Any,
    repl: t.Union[str, t.Callable[[re.Match], str]],
    ignore_case: bool = False,
    count: int = 0,
) -> str:
    """
    Replace occurrences of regex `pattern` with `repl` in `text`. Optionally, ignore case when
    replacing. Optionally, set `count` to limit number of replacements.

    Args:
        text: String to replace.
        pattern: Pattern to find and replace.
        repl: String to substitute `pattern` with.
        ignore_case: Whether to ignore case when replacing. Defaults to ``False``.
        count: Maximum number of occurrences to replace. Defaults to ``0`` which
            replaces all.

    Returns:
        Replaced string.

    Example:

        >>> reg_exp_replace('aabbcc', 'b', 'X')
        'aaXXcc'
        >>> reg_exp_replace('aabbcc', 'B', 'X', ignore_case=True)
        'aaXXcc'
        >>> reg_exp_replace('aabbcc', 'b', 'X', count=1)
        'aaXbcc'
        >>> reg_exp_replace('aabbcc', '[ab]', 'X')
        'XXXXcc'

    .. versionadded:: 3.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``re_replace`` to ``reg_exp_replace``.
    """
    if pattern is None:
        return pyd.to_string(text)

    return replace(text, pattern, repl, ignore_case=ignore_case, count=count, escape=False)


def repeat(text: t.Any, n: t.SupportsInt = 0) -> str:
    """
    Repeats the given string `n` times.

    Args:
        text: String to repeat.
        n: Number of times to repeat the string.

    Returns:
        Repeated string.

    Example:

        >>> repeat('.', 5)
        '.....'

    .. versionadded:: 1.1.0
    """
    return pyd.to_string(text) * int(n)


def replace(
    text: t.Any,
    pattern: t.Any,
    repl: t.Union[str, t.Callable[[re.Match], str]],
    ignore_case: bool = False,
    count: int = 0,
    escape: bool = True,
    from_start: bool = False,
    from_end: bool = False,
) -> str:
    """
    Replace occurrences of `pattern` with `repl` in `text`. Optionally, ignore case when replacing.
    Optionally, set `count` to limit number of replacements.

    Args:
        text: String to replace.
        pattern: Pattern to find and replace.
        repl: String to substitute `pattern` with.
        ignore_case: Whether to ignore case when replacing. Defaults to ``False``.
        count: Maximum number of occurrences to replace. Defaults to ``0`` which
            replaces all.
        escape: Whether to escape `pattern` when searching. This is needed if a
            literal replacement is desired when `pattern` may contain special regular expression
            characters. Defaults to ``True``.
        from_start: Whether to limit replacement to start of string.
        from_end: Whether to limit replacement to end of string.

    Returns:
        Replaced string.

    Example:

        >>> replace('aabbcc', 'b', 'X')
        'aaXXcc'
        >>> replace('aabbcc', 'B', 'X', ignore_case=True)
        'aaXXcc'
        >>> replace('aabbcc', 'b', 'X', count=1)
        'aaXbcc'
        >>> replace('aabbcc', '[ab]', 'X')
        'aabbcc'
        >>> replace('aabbcc', '[ab]', 'X', escape=False)
        'XXXXcc'

    .. versionadded:: 3.0.0

    .. versionchanged:: 4.1.0
        Added ``from_start`` and ``from_end`` arguments.

    .. versionchanged:: 5.0.0
        Added support for ``pattern`` as ``typing.Pattern`` object.
    """
    text = pyd.to_string(text)

    if pattern is None:
        return text

    if not pyd.is_function(repl):
        repl = pyd.to_string(repl)

    flags = re.IGNORECASE if ignore_case else 0

    if isinstance(pattern, typing.Pattern):
        pat = pattern
    else:
        pattern = pyd.to_string(pattern)

        if escape:
            pattern = re.escape(pattern)

        if from_start and not pattern.startswith("^"):
            pattern = "^" + pattern

        if from_end and not pattern.endswith("$"):
            pattern += "$"

        pat = re.compile(pattern, flags=flags)

    return pat.sub(repl, text, count=count)


def replace_end(
    text: t.Any,
    pattern: t.Any,
    repl: t.Union[str, t.Callable[[re.Match], str]],
    ignore_case: bool = False,
    escape: bool = True,
) -> str:
    """
    Like :func:`replace` except it only replaces `text` with `repl` if `pattern` mathces the end of
    `text`.

    Args:
        text: String to replace.
        pattern: Pattern to find and replace.
        repl: String to substitute `pattern` with.
        ignore_case: Whether to ignore case when replacing. Defaults to ``False``.
        escape: Whether to escape `pattern` when searching. This is needed if a
            literal replacement is desired when `pattern` may contain special regular expression
            characters. Defaults to ``True``.

    Returns:
        Replaced string.

    Example:

        >>> replace_end('aabbcc', 'b', 'X')
        'aabbcc'
        >>> replace_end('aabbcc', 'c', 'X')
        'aabbcX'

    .. versionadded:: 4.1.0
    """
    return replace(text, pattern, repl, ignore_case=ignore_case, escape=escape, from_end=True)


def replace_start(
    text: t.Any,
    pattern: t.Any,
    repl: t.Union[str, t.Callable[[re.Match], str]],
    ignore_case: bool = False,
    escape: bool = True,
) -> str:
    """
    Like :func:`replace` except it only replaces `text` with `repl` if `pattern` mathces the start
    of `text`.

    Args:
        text: String to replace.
        pattern: Pattern to find and replace.
        repl: String to substitute `pattern` with.
        ignore_case: Whether to ignore case when replacing. Defaults to ``False``.
        escape: Whether to escape `pattern` when searching. This is needed if a
            literal replacement is desired when `pattern` may contain special regular expression
            characters. Defaults to ``True``.

    Returns:
        Replaced string.

    Example:

        >>> replace_start('aabbcc', 'b', 'X')
        'aabbcc'
        >>> replace_start('aabbcc', 'a', 'X')
        'Xabbcc'

    .. versionadded:: 4.1.0
    """
    return replace(text, pattern, repl, ignore_case=ignore_case, escape=escape, from_start=True)


def separator_case(text: t.Any, separator: str) -> str:
    """
    Splits `text` on words and joins with `separator`.

    Args:
        text: String to convert.
        separator: Separator to join words with.

    Returns:
        Converted string.

    Example:

        >>> separator_case('a!!b___c.d', '-')
        'a-b-c-d'

    .. versionadded:: 3.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return separator.join(word.lower() for word in words(text) if word)


def series_phrase(
    items: t.List[t.Any],
    separator: t.Any = ", ",
    last_separator: t.Any = " and ",
    serial: bool = False,
) -> str:
    """
    Join items into a grammatical series phrase, e.g., ``"item1, item2, item3 and item4"``.

    Args:
        items: List of string items to join.
        separator: Item separator. Defaults to ``', '``.
        last_separator: Last item separator. Defaults to ``' and '``.
        serial: Whether to include `separator` with `last_separator` when number of
            items is greater than 2. Defaults to ``False``.

    Returns:
        Joined string.

    Example:

        >>> series_phrase(['apples', 'bananas', 'peaches'])
        'apples, bananas and peaches'
        >>> series_phrase(['apples', 'bananas', 'peaches'], serial=True)
        'apples, bananas, and peaches'
        >>> series_phrase(['apples', 'bananas', 'peaches'], '; ', ', or ')
        'apples; bananas, or peaches'


    .. versionadded:: 3.0.0
    """
    items = pyd.chain(items).map(pyd.to_string).compact().value()
    item_count = len(items)

    separator = pyd.to_string(separator)
    last_separator = pyd.to_string(last_separator)

    if item_count > 2 and serial:
        last_separator = separator.rstrip() + last_separator

    if item_count >= 2:
        items = items[:-2] + [last_separator.join(items[-2:])]

    return separator.join(items)


def series_phrase_serial(
    items: t.List[t.Any], separator: t.Any = ", ", last_separator: t.Any = " and "
) -> str:
    """
    Join items into a grammatical series phrase using a serial separator, e.g., ``"item1, item2,
    item3, and item4"``.

    Args:
        items: List of string items to join.
        separator: Item separator. Defaults to ``', '``.
        last_separator: Last item separator. Defaults to ``' and '``.

    Returns:
        Joined string.

    Example:

        >>> series_phrase_serial(['apples', 'bananas', 'peaches'])
        'apples, bananas, and peaches'

    .. versionadded:: 3.0.0
    """
    return series_phrase(items, separator, last_separator, serial=True)


def slugify(text: t.Any, separator: str = "-") -> str:
    """
    Convert `text` into an ASCII slug which can be used safely in URLs. Incoming `text` is converted
    to unicode and noramlzied using the ``NFKD`` form. This results in some accented characters
    being converted to their ASCII "equivalent" (e.g. ``é`` is converted to ``e``). Leading and
    trailing whitespace is trimmed and any remaining whitespace or other special characters without
    an ASCII equivalent are replaced with ``-``.

    Args:
        text: String to slugify.
        separator: Separator to use. Defaults to ``'-'``.

    Returns:
        Slugified string.

    Example:

        >>> slugify('This is a slug.') == 'this-is-a-slug'
        True
        >>> slugify('This is a slug.', '+') == 'this+is+a+slug'
        True

    .. versionadded:: 3.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.

    .. versionchanged:: 7.0.0
        Remove single quotes from output.
    """
    normalized = (
        unicodedata.normalize("NFKD", pyd.to_string(text))
        .encode("ascii", "ignore")
        .decode("utf8")
        .replace("'", "")
    )

    return separator_case(normalized, separator)


def snake_case(text: t.Any) -> str:
    """
    Converts `text` to snake case.

    Args:
        text: String to convert.

    Returns:
        String converted to snake case.

    Example:

        >>> snake_case('This is Snake Case!')
        'this_is_snake_case'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``underscore_case``.

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return "_".join(word.lower() for word in compounder(text) if word)


def split(text: t.Any, separator: t.Union[str, Unset, None] = UNSET) -> t.List[str]:
    """
    Splits `text` on `separator`. If `separator` not provided, then `text` is split on whitespace.
    If `separator` is falsey, then `text` is split on every character.

    Args:
        text: String to explode.
        separator: Separator string to split on. Defaults to ``NoValue``.

    Returns:
        Split string.

    Example:

        >>> split('one potato, two potatoes, three potatoes, four!')
        ['one', 'potato,', 'two', 'potatoes,', 'three', 'potatoes,', 'four!']
        >>> split('one potato, two potatoes, three potatoes, four!', ',')
        ['one potato', ' two potatoes', ' three potatoes', ' four!']

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Changed `separator` default to ``NoValue`` and supported splitting on whitespace by default.

    .. versionchanged:: 4.0.0
        Removed alias ``explode``.
    """
    text = pyd.to_string(text)

    if separator is UNSET:
        ret = text.split()
    elif separator:
        ret = text.split(separator)
    else:
        ret = chars(text)

    return ret


def start_case(text: t.Any) -> str:
    """
    Convert `text` to start case.

    Args:
        text: String to convert.

    Returns:
        String converted to start case.

    Example:

        >>> start_case("fooBar")
        'Foo Bar'

    .. versionadded:: 3.1.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return " ".join(capitalize(word, strict=False) for word in compounder(text))


def starts_with(text: t.Any, target: t.Any, position: int = 0) -> bool:
    """
    Checks if `text` starts with a given target string.

    Args:
        text: String to check.
        target: String to check for.
        position: Position to search from. Defaults to beginning of `text`.

    Returns:
        Whether `text` starts with `target`.

    Example:

        >>> starts_with('abcdef', 'a')
        True
        >>> starts_with('abcdef', 'b')
        False
        >>> starts_with('abcdef', 'a', 1)
        False

    .. versionadded:: 1.1.0
    """
    text = pyd.to_string(text)
    target = pyd.to_string(target)
    return text[position:].startswith(target)


def strip_tags(text: t.Any) -> str:
    """
    Removes all HTML tags from `text`.

    Args:
        text: String to strip.

    Returns:
        String without HTML tags.

    Example:

        >>> strip_tags('<a href="#">Some link</a>')
        'Some link'

    .. versionadded:: 3.0.0
    """
    return RE_HTML_TAGS.sub("", pyd.to_string(text))


def substr_left(text: t.Any, subtext: str) -> str:
    """
    Searches `text` from left-to-right for `subtext` and returns a substring consisting of the
    characters in `text` that are to the left of `subtext` or all string if no match found.

    Args:
        text: String to partition.
        subtext: String to search for.

    Returns:
        Substring to left of `subtext`.

    Example:

        >>> substr_left('abcdefcdg', 'cd')
        'ab'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.partition(subtext)[0] if subtext else text


def substr_left_end(text: t.Any, subtext: str) -> str:
    """
    Searches `text` from right-to-left for `subtext` and returns a substring consisting of the
    characters in `text` that are to the left of `subtext` or all string if no match found.

    Args:
        text: String to partition.
        subtext: String to search for.

    Returns:
        Substring to left of `subtext`.

    Example:

        >>> substr_left_end('abcdefcdg', 'cd')
        'abcdef'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.rpartition(subtext)[0] or text if subtext else text


def substr_right(text: t.Any, subtext: str) -> str:
    """
    Searches `text` from right-to-left for `subtext` and returns a substring consisting of the
    characters in `text` that are to the right of `subtext` or all string if no match found.

    Args:
        text: String to partition.
        subtext: String to search for.

    Returns:
        Substring to right of `subtext`.

    Example:

        >>> substr_right('abcdefcdg', 'cd')
        'efcdg'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.partition(subtext)[2] or text if subtext else text


def substr_right_end(text: t.Any, subtext: str) -> str:
    """
    Searches `text` from left-to-right for `subtext` and returns a substring consisting of the
    characters in `text` that are to the right of `subtext` or all string if no match found.

    Args:
        text: String to partition.
        subtext: String to search for.

    Returns:
        Substring to right of `subtext`.

    Example:

        >>> substr_right_end('abcdefcdg', 'cd')
        'g'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.rpartition(subtext)[2] if subtext else text


def successor(char: t.Any) -> str:
    """
    Return the successor character of `char`.

    Args:
        char: Character to find the successor of.

    Returns:
        Successor character.

    Example:

        >>> successor('b')
        'c'
        >>> successor('B')
        'C'
        >>> successor('2')
        '3'

    .. versionadded:: 3.0.0
    """
    char = pyd.to_string(char)
    return chr(ord(char) + 1)


def surround(text: t.Any, wrapper: t.Any) -> str:
    """
    Surround a string with another string.

    Args:
        text: String to surround with `wrapper`.
        wrapper: String by which `text` is to be surrounded.

    Returns:
        Surrounded string.

    Example:

        >>> surround('abc', '"')
        '"abc"'
        >>> surround('abc', '!')
        '!abc!'

    .. versionadded:: 2.4.0
    """
    text = pyd.to_string(text)
    wrapper = pyd.to_string(wrapper)
    return f"{wrapper}{text}{wrapper}"


def swap_case(text: t.Any) -> str:
    """
    Swap case of `text` characters.

    Args:
        text: String to swap case.

    Returns:
        String with swapped case.

    Example:

        >>> swap_case('aBcDeF')
        'AbCdEf'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    return text.swapcase()


def title_case(text: t.Any) -> str:
    """
    Convert `text` to title case.

    Args:
        text: String to convert.

    Returns:
        String converted to title case.

    Example:

        >>> title_case("bob's shop")
        "Bob's Shop"

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    # NOTE: Can't use text.title() since it doesn't handle apostrophes.
    return " ".join(word.capitalize() for word in re.split(" ", text))


def to_lower(text: t.Any) -> str:
    """
    Converts the given :attr:`text` to lower text.

    Args:
        text: String to convert.

    Returns:
        String converted to lower case.

    Example:

        >>> to_lower('--Foo-Bar--')
        '--foo-bar--'
        >>> to_lower('fooBar')
        'foobar'
        >>> to_lower('__FOO_BAR__')
        '__foo_bar__'

    .. versionadded:: 4.0.0
    """
    return pyd.to_string(text).lower()


def to_upper(text: t.Any) -> str:
    """
    Converts the given :attr:`text` to upper text.

    Args:
        text: String to convert.

    Returns:
        String converted to upper case.

    Example:

        >>> to_upper('--Foo-Bar--')
        '--FOO-BAR--'
        >>> to_upper('fooBar')
        'FOOBAR'
        >>> to_upper('__FOO_BAR__')
        '__FOO_BAR__'

    .. versionadded:: 4.0.0
    """
    return pyd.to_string(text).upper()


def trim(text: t.Any, chars: t.Union[str, None] = None) -> str:
    r"""
    Removes leading and trailing whitespace or specified characters from `text`.

    Args:
        text: String to trim.
        chars: Specific characters to remove.

    Returns:
        Trimmed string.

    Example:

        >>> trim('  abc efg\r\n ')
        'abc efg'

    .. versionadded:: 1.1.0
    """
    # pylint: disable=redefined-outer-name
    text = pyd.to_string(text)
    return text.strip(chars)


def trim_end(text: t.Any, chars: t.Union[str, None] = None) -> str:
    r"""
    Removes trailing whitespace or specified characters from `text`.

    Args:
        text: String to trim.
        chars: Specific characters to remove.

    Returns:
        Trimmed string.

    Example:

        >>> trim_end('  abc efg\r\n ')
        '  abc efg'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Renamed from ``trim_right`` to ``trim_end``.
    """
    text = pyd.to_string(text)
    return text.rstrip(chars)


def trim_start(text: t.Any, chars: t.Union[str, None] = None) -> str:
    r"""
    Removes leading  whitespace or specified characters from `text`.

    Args:
        text: String to trim.
        chars: Specific characters to remove.

    Returns:
        Trimmed string.

    Example:

        >>> trim_start('  abc efg\r\n ')
        'abc efg\r\n '

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Renamed from ``trim_left`` to ``trim_start``.
    """
    text = pyd.to_string(text)
    return text.lstrip(chars)


def truncate(
    text: t.Any,
    length: int = 30,
    omission: str = "...",
    separator: t.Union[str, re.Pattern, None] = None,
) -> str:
    """
    Truncates `text` if it is longer than the given maximum string length. The last characters of
    the truncated string are replaced with the omission string which defaults to ``...``.

    Args:
        text: String to truncate.
        length: Maximum string length. Defaults to ``30``.
        omission: String to indicate text is omitted.
        separator: Separator pattern to truncate to.

    Returns:
        Truncated string.

    Example:

        >>> truncate('hello world', 5)
        'he...'
        >>> truncate('hello world', 5, '..')
        'hel..'
        >>> truncate('hello world', 10)
        'hello w...'
        >>> truncate('hello world', 10, separator=' ')
        'hello...'

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``trunc``.
    """
    text = pyd.to_string(text)

    if len(text) <= length:
        return text

    omission_len = len(omission)
    text_len = length - omission_len
    text = text[:text_len]

    trunc_len = len(text)

    if pyd.is_string(separator):
        trunc_len = text.rfind(separator)
    elif pyd.is_reg_exp(separator):
        last = None
        for match in separator.finditer(text):
            last = match

        if last is not None:
            trunc_len = last.start()

    return text[:trunc_len] + omission


def unescape(text: t.Any) -> str:
    """
    The inverse of :func:`escape`. This method converts the HTML entities ``&amp;``, ``&lt;``,
    ``&gt;``, ``&quot;``, ``&#39;``, and ``&#96;`` in `text` to their corresponding characters.

    Args:
        text: String to unescape.

    Returns:
        HTML unescaped string.

    Example:

        >>> results = unescape('&quot;1 &gt; 2 &amp;&amp; 3 &lt; 4&quot;')
        >>> results == '"1 > 2 && 3 < 4"'
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Moved to :mod:`pydash.strings`.
    """
    text = pyd.to_string(text)
    return html.unescape(text)


def upper_case(text: t.Any) -> str:
    """
    Converts string to upper case, as space separated words.

    Args:
        text: String to be converted to uppercase.

    Returns:
        String converted to uppercase, as space separated words.

    Example:

        >>> upper_case('--foo-bar--')
        'FOO BAR'
        >>> upper_case('fooBar')
        'FOO BAR'
        >>> upper_case('/?*Foo10/;"B*Ar')
        'FOO 10 B AR'

    .. versionadded:: 4.0.0

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    return " ".join(compounder(text)).upper()


def upper_first(text: str) -> str:
    """
    Converts the first character of string to upper case.

    Args:
        text: String passed in by the user.

    Returns:
        String in which the first character is converted to upper case.

    Example:

        >>> upper_first('fred')
        'Fred'
        >>> upper_first('foo bar')
        'Foo bar'
        >>> upper_first('1foobar')
        '1foobar'
        >>> upper_first(';foobar')
        ';foobar'

    .. versionadded:: 4.0.0
    """
    return text[:1].upper() + text[1:]


def unquote(text: t.Any, quote_char: t.Any = '"') -> str:
    """
    Unquote `text` by removing `quote_char` if `text` begins and ends with it.

    Args:
        text: String to unquote.
        quote_char: Quote character to remove. Defaults to `"`.

    Returns:
        Unquoted string.

    Example:

        >>> unquote('"abc"')
        'abc'
        >>> unquote('"abc"', '#')
        '"abc"'
        >>> unquote('#abc', '#')
        '#abc'
        >>> unquote('#abc#', '#')
        'abc'

    .. versionadded:: 3.0.0
    """
    text = pyd.to_string(text)
    inner = text[1:-1]

    if text == f"{quote_char}{inner}{quote_char}":
        text = inner

    return text


def url(*paths: t.Any, **params: t.Any) -> str:
    """
    Combines a series of URL paths into a single URL. Optionally, pass in keyword arguments to
    append query parameters.

    Args:
        paths: URL paths to combine.

    Keyword Args:
        params: Query parameters.

    Returns:
        URL string.

    Example:

        >>> link = url('a', 'b', ['c', 'd'], '/', q='X', y='Z')
        >>> path, params = link.split('?')
        >>> path == 'a/b/c/d/'
        True
        >>> set(params.split('&')) == set(['q=X', 'y=Z'])
        True

    .. versionadded:: 2.2.0
    """
    # allow reassignment different type
    paths = pyd.chain(paths).flatten_deep().map(pyd.to_string).value()  # type: ignore
    paths_list = []
    params_list = flatten_url_params(params)

    for path in paths:
        scheme, netloc, path, query, fragment = urlsplit(path)
        query = parse_qsl(query)
        params_list += query
        paths_list.append(urlunsplit((scheme, netloc, path, "", fragment)))

    path = delimitedpathjoin("/", *paths_list)
    scheme, netloc, path, query, fragment = urlsplit(path)
    query = urlencode(params_list)

    return urlunsplit((scheme, netloc, path, query, fragment))


def words(text: t.Any, pattern: t.Union[str, None] = None) -> t.List[str]:
    """
    Return list of words contained in `text`.

    References:
        https://github.com/lodash/lodash/blob/master/words.js#L30

    Args:
        text: String to split.
        pattern: Custom pattern to split words on. Defaults to ``None``.

    Returns:
        List of words.

    Example:

        >>> words('a b, c; d-e')
        ['a', 'b', 'c', 'd', 'e']
        >>> words('fred, barney, & pebbles', '/[^, ]+/g')
        ['fred', 'barney', '&', 'pebbles']

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.2.0
        Added `pattern` argument.

    .. versionchanged:: 3.2.0
        Improved matching for one character words.

    .. versionchanged:: 5.0.0
        Improved unicode word support.
    """
    text = pyd.to_string(text)
    if pattern is None:
        if has_unicode_word(text):
            reg_exp = JS_RE_UNICODE_WORDS
        else:
            reg_exp = JS_RE_ASCII_WORDS
    else:
        reg_exp = JSRegExp(pattern)
    return reg_exp.find(text)


#
# Utility functions not a part of main API
#


def compounder(text):
    """
    Remove single quote before passing into words() to match Lodash-style outputs.

    Required by certain functions such as kebab_case, camel_case, start_case etc.

    References:
        https://github.com/lodash/lodash/blob/4.17.15/lodash.js#L4968
    """
    return words(deburr(RE_APOS.sub("", pyd.to_string(text))))


def has_unicode_word(text):
    """
    Check if the text contains unicode or requires more complex regex to handle.

    References:
        https://github.com/lodash/lodash/blob/master/words.js#L3
    """
    result = RE_HAS_UNICODE_WORD.search(text)
    return bool(result)


def delimitedpathjoin(delimiter, *paths):
    """
    Join delimited path using specified delimiter.

    >>> assert delimitedpathjoin('.', '') == ''
    >>> assert delimitedpathjoin('.', '.') == '.'
    >>> assert delimitedpathjoin('.', ['', '.a']) == '.a'
    >>> assert delimitedpathjoin('.', ['a', '.']) == 'a.'
    >>> assert delimitedpathjoin('.', ['', '.a', '', '', 'b']) == '.a.b'
    >>> ret = '.a.b.c.d.e.'
    >>> assert delimitedpathjoin('.', ['.a.', 'b.', '.c', 'd', 'e.']) == ret
    >>> assert delimitedpathjoin('.', ['a', 'b', 'c']) == 'a.b.c'
    >>> ret = 'a.b.c.d.e.f'
    >>> assert delimitedpathjoin('.', ['a.b', '.c.d.', '.e.f']) == ret
    >>> ret = '.a.b.c.1.'
    >>> assert delimitedpathjoin('.', '.', 'a', 'b', 'c', 1, '.') == ret
    >>> assert delimitedpathjoin('.', []) == ''
    """
    paths = [pyd.to_string(path) for path in pyd.flatten_deep(paths) if path]

    if len(paths) == 1:
        # Special case where there's no need to join anything. Doing this because if
        # path==[delimiter], then an extra delimiter would be added if the else clause ran instead.
        path = paths[0]
    else:
        leading = delimiter if paths and paths[0].startswith(delimiter) else ""
        trailing = delimiter if paths and paths[-1].endswith(delimiter) else ""
        middle = delimiter.join([path.strip(delimiter) for path in paths if path.strip(delimiter)])
        path = "".join([leading, middle, trailing])

    return path


def flatten_url_params(
    params: t.Union[
        t.Dict[T, t.Union[T2, t.Iterable[T2]]],
        t.List[t.Tuple[T, t.Union[T2, t.Iterable[T2]]]],
    ],
) -> t.List[t.Tuple[T, T2]]:
    """
    Flatten URL params into list of tuples. If any param value is a list or tuple, then map each
    value to the param key.

    >>> params = [('a', 1), ('a', [2, 3])]
    >>> assert flatten_url_params(params) == [('a', 1), ('a', 2), ('a', 3)]
    >>> params = {'a': [1, 2, 3]}
    >>> assert flatten_url_params(params) == [('a', 1), ('a', 2), ('a', 3)]
    """
    if isinstance(params, dict):
        params = list(params.items())

    flattened: t.List = []
    for param, value in params:
        if isinstance(value, (list, tuple)):
            flattened += zip([param] * len(value), value)
        else:
            flattened.append((param, value))

    return flattened



import typing as t

import pytest

import pydash as _


@pytest.mark.mypy_testing
def test_mypy_after() -> None:
    def func(a: int, b: t.Dict[str, int], c: bytes) -> str:
        return f"{a} {b} {c!r}"

    after_func = _.after(func, 1)
    reveal_type(after_func)  # R: pydash.functions.After[[a: builtins.int, b: builtins.dict[builtins.str, builtins.int], c: builtins.bytes], builtins.str]
    reveal_type(after_func(1, {}, b""))  # R: Union[builtins.str, None]


@pytest.mark.mypy_testing
def test_mypy_ary() -> None:
    def func(a: int, b: int, c: int = 0, d: int = 5) -> t.Tuple[int, int, int, int]:
        return (a, b, c, d)

    ary_func = _.ary(func, 2)
    reveal_type(ary_func(1, 2, 3, 4, 5, 6))  # R: Tuple[builtins.int, builtins.int, builtins.int, builtins.int]
    reveal_type(ary_func(1, 2, 3, 4, 5, 6, c=10, d=20))  # R: Tuple[builtins.int, builtins.int, builtins.int, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_before() -> None:
    def func(a: int, b: int, c: int) -> t.Tuple[int, int, int]:
        return (a, b, c)

    before_func = _.before(func, 3)
    reveal_type(before_func(1, 2, 3))  # R: Union[Tuple[builtins.int, builtins.int, builtins.int], None]


@pytest.mark.mypy_testing
def test_mypy_conjoin() -> None:
    def up_3(x: int) -> bool:
        return x > 3
    def is_int(x: t.Any) -> int:
        return isinstance(x, int)
    conjoiner = _.conjoin(is_int, up_3)

    reveal_type(conjoiner([1, 2, 3]))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_curry() -> None:
    def func(a: int, b: str, c: bytes) -> t.Tuple[int, str, bytes]:
        return (a, b, c)

    currier = _.curry(func)
    reveal_type(currier)  # R: pydash.functions.CurryThree[builtins.int, builtins.str, builtins.bytes, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier1 = currier(1)
    reveal_type(currier1)  # R: pydash.functions.CurryTwo[builtins.str, builtins.bytes, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier2 = currier1("hi")
    reveal_type(currier2)  # R: pydash.functions.CurryOne[builtins.bytes, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier3 = currier2(b"hi again")
    reveal_type(currier3)  # R: Tuple[builtins.int, builtins.str, builtins.bytes]


@pytest.mark.mypy_testing
def test_mypy_curry_right() -> None:
    def func(a: int, b: str, c: bytes) -> t.Tuple[int, str, bytes]:
        return (a, b, c)

    currier = _.curry_right(func)
    reveal_type(currier)  # R: pydash.functions.CurryRightThree[builtins.bytes, builtins.str, builtins.int, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier1 = currier(b"hi again")
    reveal_type(currier1)  # R: pydash.functions.CurryRightTwo[builtins.str, builtins.int, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier2 = currier1("hi")
    reveal_type(currier2)  # R: pydash.functions.CurryRightOne[builtins.int, Tuple[builtins.int, builtins.str, builtins.bytes]]

    currier3 = currier2(1)
    reveal_type(currier3)  # R: Tuple[builtins.int, builtins.str, builtins.bytes]


@pytest.mark.mypy_testing
def test_mypy_debounce() -> None:
    def func(a: int, b: str) -> t.Tuple[int, str]:
        return (a, b)

    debounced = _.debounce(func, 5000)

    reveal_type(debounced)  # R: pydash.functions.Debounce[[a: builtins.int, b: builtins.str], Tuple[builtins.int, builtins.str]]
    reveal_type(debounced(5, "hi"))  # R: Tuple[builtins.int, builtins.str]


@pytest.mark.mypy_testing
def test_mypy_delay() -> None:
    def func(a: int, b: str) -> t.Tuple[int, str]:
        return (a, b)

    reveal_type(_.delay(func, 0, 5, "hi"))  # R: Tuple[builtins.int, builtins.str]


@pytest.mark.mypy_testing
def test_mypy_disjoin() -> None:
    def is_float(x: t.Any) -> bool:
        return isinstance(x, float)
    def is_int(x: t.Any) -> bool:
        return isinstance(x, int)
    disjoiner = _.disjoin(is_float, is_int)

    reveal_type(disjoiner)  # R: pydash.functions.Disjoin[Any]
    reveal_type(disjoiner([1, '2', '3']))  # R: builtins.bool
    reveal_type(disjoiner([1.0, '2', '3']))  # R: builtins.bool
    reveal_type(disjoiner(['1', '2', '3']))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_flip() -> None:
    def func(a: int, b: str, c: bytes) -> t.Tuple[int, str, bytes]:
        return (a, b, c)

    reveal_type(_.flip(func))  # R: def (builtins.bytes, builtins.str, builtins.int) -> Tuple[builtins.int, builtins.str, builtins.bytes]


@pytest.mark.mypy_testing
def test_mypy_flow() -> None:
    def mult_5(x: int) -> int:
        return x * 5
    def div_10(x: int) -> float:
        return x / 10.0
    def pow_2(x: float) -> float:
        return x ** 2
    def sum_list(x: t.List[int]) -> int:
        return sum(x)

    ops = _.flow(sum_list, mult_5, div_10, pow_2)
    reveal_type(ops)  # R: pydash.functions.Flow[[x: builtins.list[builtins.int]], builtins.float]
    reveal_type(ops([1, 2, 3, 4]))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_flow_right() -> None:
    def mult_5(x: float) -> float:
        return x * 5
    def div_10(x: int) -> float:
        return x / 10
    def pow_2(x: int) -> int:
        return x ** 2
    def sum_list(x: t.List[int]) -> int:
        return sum(x)

    ops = _.flow_right(mult_5, div_10, pow_2, sum_list)
    reveal_type(ops)  # R: pydash.functions.Flow[[x: builtins.list[builtins.int]], builtins.float]
    reveal_type(ops([1, 2, 3, 4]))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_iterated() -> None:
    def double(x: int) -> int:
        return x * 2

    doubler = _.iterated(double)
    reveal_type(doubler)  # R: pydash.functions.Iterated[builtins.int]

    reveal_type(doubler(4, 5))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_juxtapose() -> None:
    def double(x: int) -> int:
        return x * 2
    def triple(x: int) -> int:
        return x * 3
    def quadruple(x: int) -> int:
        return x * 4

    f = _.juxtapose(double, triple, quadruple)
    reveal_type(f)  # R: pydash.functions.Juxtapose[[x: builtins.int], builtins.int]
    reveal_type(f(5))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_negate() -> None:
    def is_number(x: t.Any) -> bool:
        return isinstance(x, (int, float))

    not_is_number = _.negate(is_number)
    reveal_type(not_is_number)  # R: pydash.functions.Negate[[x: Any]]
    reveal_type(not_is_number('1'))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_once() -> None:
    def first_arg(*args: int) -> int:
        return args[0]

    oncer = _.once(first_arg)
    reveal_type(oncer)  # R: pydash.functions.Once[[*args: builtins.int], builtins.int]
    reveal_type(oncer(6))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_over_args() -> None:
    def squared(x: int) -> int:
        return x ** 2
    def double(x: int) -> int:
        return x * 2
    def in_list(x: int, y: int) -> t.List[int]:
        return [x, y]

    modder = _.over_args(in_list, squared, double)
    reveal_type(modder)  # R: def (builtins.int, builtins.int) -> builtins.list[builtins.int]
    reveal_type(modder(5, 10))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_partial() -> None:
    def cut(array: t.List[int], n: int) -> t.List[int]:
        return array[n:]

    dropper = _.partial(cut, [1, 2, 3, 4])
    reveal_type(dropper)  # R: pydash.functions.Partial[builtins.list[builtins.int]]
    reveal_type(dropper(2))  # R: builtins.list[builtins.int]

    myrest = _.partial(cut, n=1)
    reveal_type(myrest)  # R: pydash.functions.Partial[builtins.list[builtins.int]]
    reveal_type(myrest([1, 2, 3, 4]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_partial_right() -> None:
    def cut(array: t.List[int], n: int) -> t.List[int]:
        return array[n:]

    myrest = _.partial_right(cut, 1)
    reveal_type(myrest)  # R: pydash.functions.Partial[builtins.list[builtins.int]]
    reveal_type(myrest([1, 2, 3, 4]))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_rearg() -> None:
    def func(x: int, y: int) -> t.List[int]:
        return [x, y]

    jumble = _.rearg(func, 1, 2, 3)
    reveal_type(jumble)  # R: pydash.functions.Rearg[[x: builtins.int, y: builtins.int], builtins.list[builtins.int]]
    reveal_type(jumble(1, 2))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_spread() -> None:
    def greet_people(*people: str) -> str:
        return 'Hello ' + ', '.join(people) + '!'

    greet = _.spread(greet_people)
    reveal_type(greet)  # R: pydash.functions.Spread[builtins.str]
    reveal_type(greet(['Mike', 'Don', 'Leo']))  # R: builtins.str

@pytest.mark.mypy_testing
def test_mypy_throttle() -> None:
    def func(x: int) -> int:
        return x

    throttled = _.throttle(func, 0)
    reveal_type(throttled)  # R: pydash.functions.Throttle[[x: builtins.int], builtins.int]
    reveal_type(throttled(5))  # R: builtins.int

@pytest.mark.mypy_testing
def test_mypy_unary() -> None:
    def func(a: int, b: int = 1, c: int = 0, d: int = 5) -> t.Tuple[int, int, int, int]:
        return (a, b, c, d)

    unary_func = _.unary(func)
    reveal_type(unary_func)  # R: pydash.functions.Ary[Tuple[builtins.int, builtins.int, builtins.int, builtins.int]]
    reveal_type(unary_func(1, 2, 3, 4, 5, 6))  # R: Tuple[builtins.int, builtins.int, builtins.int, builtins.int]
    reveal_type(unary_func(1, 2, 3, 4, 5, 6, c=10, d=20))  # R: Tuple[builtins.int, builtins.int, builtins.int, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_wrap() -> None:
    def as_tuple(x: str, y: int) -> t.Tuple[str, int]:
        return (x, y)

    wrapper = _.wrap('hello', as_tuple)
    reveal_type(wrapper)  # R: pydash.functions.Partial[Tuple[builtins.str, builtins.int]]
    reveal_type(wrapper(1))  # R: Tuple[builtins.str, builtins.int]



import typing as t

import pytest

import pydash as _


@pytest.mark.mypy_testing
def test_mypy_attempt() -> None:
    reveal_type(_.attempt(lambda x: x/0, 1))  # R: Union[Any, builtins.Exception]


@pytest.mark.mypy_testing
def test_mypy_cond() -> None:
    def is_1(n: int) -> bool:
        return n == 1
    func = _.cond(
        [
            (is_1, _.constant("matches 1")),
        ]
    )
    reveal_type(func)  # R: def (n: builtins.int) -> builtins.str


@pytest.mark.mypy_testing
def test_mypy_conforms() -> None:
    def higher_than_1(n: int) -> bool:
        return n > 1
    def is_0(n: int) -> bool:
        return n == 0

    reveal_type(_.conforms({'b': higher_than_1}))  # R: def (builtins.dict[builtins.str, builtins.int]) -> builtins.bool
    reveal_type(_.conforms([higher_than_1, is_0]))  # R: def (builtins.list[builtins.int]) -> builtins.bool


@pytest.mark.mypy_testing
def test_mypy_conforms_to() -> None:
    reveal_type(_.conforms_to({'b': 2}, {'b': lambda n: n > 1}))  # R: builtins.bool
    reveal_type(_.conforms_to({'b': 0}, {'b': lambda n: n > 1}))  # R: builtins.bool
    reveal_type(_.conforms_to([2, 0], [lambda n: n > 1, lambda n: n == 0]))  # R: builtins.bool
    reveal_type(_.conforms_to([0, 0], [lambda n: n > 1, lambda n: n == 0]))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_constant() -> None:
    reveal_type(_.constant(3.14))  # R: def (*Any, **Any) -> builtins.float
    reveal_type(_.constant("hello"))  # R: def (*Any, **Any) -> builtins.str


@pytest.mark.mypy_testing
def test_mypy_default_to() -> None:
    reveal_type(_.default_to(1, 10))  # R: builtins.int
    reveal_type(_.default_to(None, 10))  # R: builtins.int
    reveal_type(_.default_to(None, None))  # R: None


@pytest.mark.mypy_testing
def test_mypy_default_to_any() -> None:
    n1, n2, n3 = 1, 10, 20
    reveal_type(_.default_to_any(n1, n2, n3))  # R: builtins.int
    reveal_type(_.default_to_any(None, n2, n3))  # R: builtins.int
    reveal_type(_.default_to_any(None, None, n3))  # R: builtins.int
    reveal_type(_.default_to_any(None, None, None))  # R: None


@pytest.mark.mypy_testing
def test_mypy_identity() -> None:
    reveal_type(_.identity(1))  # R: builtins.int
    reveal_type(_.identity(1, 2, 3))  # R: builtins.int
    reveal_type(_.identity())  # R: None


@pytest.mark.mypy_testing
def test_mypy_iteratee() -> None:
    def add(x: int, y: int) -> int:
        return x + y

    reveal_type(_.iteratee('data'))  # R: def (*Any, **Any) -> Any
    reveal_type(_.iteratee({'active': True})) # R: def (*Any, **Any) -> Any
    reveal_type(_.iteratee(add))  # R: def (x: builtins.int, y: builtins.int) -> builtins.int


@pytest.mark.mypy_testing
def test_mypy_matches() -> None:
    reveal_type(_.matches({'a': {'b': 2}}))  # R: def (Any) -> builtins.bool


@pytest.mark.mypy_testing
def test_mypy_matches_property() -> None:
    reveal_type(_.matches_property('a', 1))  # R: def (Any) -> builtins.bool
    reveal_type(_.matches_property(0, 1))  # R: def (Any) -> builtins.bool
    reveal_type(_.matches_property('a', 2))  # R: def (Any) -> builtins.bool


@pytest.mark.mypy_testing
def test_mypy_memoize() -> None:
    def add(x: int, y: int) -> int:
        return x + y

    def mul(x: int, y: int) -> int:
        return x * y

    memoized_add = _.memoize(add)
    reveal_type(memoized_add)  # R: pydash.utilities.MemoizedFunc[[x: builtins.int, y: builtins.int], builtins.int, builtins.str]
    reveal_type(memoized_add.cache)  # R: builtins.dict[builtins.str, builtins.int]

    memoized_add_resolv = _.memoize(add, mul)
    reveal_type(memoized_add_resolv)  # R: pydash.utilities.MemoizedFunc[[x: builtins.int, y: builtins.int], builtins.int, builtins.int]
    reveal_type(memoized_add_resolv.cache)  # R: builtins.dict[builtins.int, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_method() -> None:
    obj = {'a': {'b': [None, lambda x: x]}}
    reveal_type(_.method('a.b.1'))  # R: def (*Any, **Any) -> Any


@pytest.mark.mypy_testing
def test_mypy_method_of() -> None:
    obj = {'a': {'b': [None, lambda x: x]}}
    reveal_type(_.method_of(obj))  # R: def (*Any, **Any) -> Any


@pytest.mark.mypy_testing
def test_mypy_noop() -> None:
    reveal_type(_.noop(5, "hello", {}))  # R: None


@pytest.mark.mypy_testing
def test_mypy_nth_arg() -> None:
    reveal_type(_.nth_arg(1))  # R: def (*Any, **Any) -> Any


@pytest.mark.mypy_testing
def test_mypy_now() -> None:
    reveal_type(_.now())  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_over() -> None:
    def first(*args: int) -> int:
        return args[0]
    def second(*args: int) -> int:
        return args[1]

    reveal_type(_.over([first, second]))  # R: def (*args: builtins.int) -> builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_over_every() -> None:
    def is_two(x: int) -> bool:
        return x == 2
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.over_every([is_one, is_two]))  # R: def (x: builtins.int) -> builtins.bool


@pytest.mark.mypy_testing
def test_mypy_over_some() -> None:
    def is_two(x: int) -> bool:
        return x == 2
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.over_some([is_one, is_two]))  # R: def (x: builtins.int) -> builtins.bool


@pytest.mark.mypy_testing
def test_mypy_property_() -> None:
    reveal_type(_.property_('data'))  # R: def (Any) -> Any
    reveal_type(_.property_(0))  # R: def (Any) -> Any


@pytest.mark.mypy_testing
def test_mypy_properties() -> None:
    reveal_type(_.properties('a', 'b', ['c', 'd', 'e']))  # R: def (Any) -> Any


@pytest.mark.mypy_testing
def test_mypy_property_of() -> None:
    reveal_type(_.property_of({'a': 1, 'b': 2, 'c': 3}))  # R: def (Union[typing.Hashable, builtins.list[typing.Hashable]]) -> Any


@pytest.mark.mypy_testing
def test_mypy_random() -> None:
    reveal_type(_.random())  # R: builtins.int
    reveal_type(_.random(5, 10))  # R: builtins.int
    reveal_type(_.random(floating=True))  # R: builtins.float
    reveal_type(_.random(5.4, 10))  # R: builtins.float
    reveal_type(_.random(5, 10.0))  # R: builtins.float
    reveal_type(_.random(5.5, 10.0))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_range_() -> None:
    reveal_type(list(_.range_(5)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_(1, 4)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_(0, 6, 2)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_(4, 1)))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_range_right() -> None:
    reveal_type(list(_.range_right(5)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_right(1, 4)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_right(0, 6, 2)))  # R: builtins.list[builtins.int]
    reveal_type(list(_.range_right(4, 1)))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_result() -> None:
    reveal_type(_.result({'a': 1, 'b': lambda: 2}, 'a'))  # R: Any
    reveal_type(_.result(None, 'a', None))  # R: None
    reveal_type(_.result(None, 'a', 5))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_retry() -> None:
    def add(x: int, y: int) -> int:
        return x + y

    r = _.retry(attempts=3, delay=0)
    reveal_type(r(add))  # R: def (x: builtins.int, y: builtins.int) -> builtins.int


@pytest.mark.mypy_testing
def test_mypy_stub_list() -> None:
    reveal_type(_.stub_list())  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_stub_dict() -> None:
    reveal_type(_.stub_dict())  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_stub_false() -> None:
    reveal_type(_.stub_false())  # R: Literal[False]


@pytest.mark.mypy_testing
def test_mypy_stub_string() -> None:
    reveal_type(_.stub_string())  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_stub_true() -> None:
    reveal_type(_.stub_true())  # R: Literal[True]


@pytest.mark.mypy_testing
def test_mypy_times() -> None:
    def to_str(x: int) -> str:
        return str(x)

    reveal_type(_.times(5))  # R: builtins.list[builtins.int]
    reveal_type(_.times(5, to_str))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_to_path() -> None:
    reveal_type(_.to_path('a.b.c'))  # R: builtins.list[typing.Hashable]


@pytest.mark.mypy_testing
def test_mypy_unique_id() -> None:
    reveal_type(_.unique_id())  # R: builtins.str
    reveal_type(_.unique_id('id_'))  # R: builtins.str



import typing as t

from typing_extensions import Protocol


if t.TYPE_CHECKING:
    # F401: imported but unused
    # py3.7 flake8 doesn't get it for type aliases
    from decimal import Decimal  # noqa: F401

IterateeObjT = t.Union[int, str, t.List, t.Tuple, t.Dict]
NumberT = t.Union[float, int, "Decimal"]
NumberNoDecimalT = t.Union[float, int]
PathT = t.Union[t.Hashable, t.List[t.Hashable]]


_T_co = t.TypeVar("_T_co", covariant=True)
_T_contra = t.TypeVar("_T_contra", contravariant=True)


class SupportsMul(Protocol[_T_contra, _T_co]):
    def __mul__(self, x: _T_contra) -> _T_co:
        ...


class SupportsRound(Protocol[_T_co]):
    def __round__(self) -> _T_co:
        ...



"""
This module provides the CLI interface for invoke tasks.

All tasks can be executed from this file's directory using:

    $ inv <task>

Where <task> is a function defined below with the @task decorator.
"""

from functools import partial
import os
from pathlib import Path
import sys
import tempfile
import typing as t

from invoke import Context, Exit, UnexpectedExit, run as _run, task


PACKAGE_NAME = "pydash"
PACKAGE_SOURCE = f"src/{PACKAGE_NAME}"
MYPY_TESTS_DIR = "tests/pytest_mypy_testing"
TEST_TARGETS = f"{PACKAGE_SOURCE} tests"
LINT_TARGETS = f"{TEST_TARGETS} tasks.py"
LINT_EXCLUDE = MYPY_TESTS_DIR
EXIT_EXCEPTIONS = (Exit, UnexpectedExit, SystemExit)


# Set pyt=True to enable colored output when available.
run = partial(_run, pty=True)


@task()
def black(ctx: Context, target: t.Optional[str] = None, quiet: bool = False) -> None:
    """Autoformat code using black."""
    run(f"black --exclude='{LINT_EXCLUDE}' {target or LINT_TARGETS}", hide=quiet)


@task()
def isort(ctx: Context, target: t.Optional[str] = None, quiet: bool = False) -> None:
    """Autoformat Python imports."""
    run(f"isort {target or LINT_TARGETS}", hide=quiet)


@task()
def docformatter(ctx: Context, target: t.Optional[str] = None) -> None:
    """Autoformat docstrings using docformatter."""
    run(
        f"docformatter -r {target or LINT_TARGETS} "
        f"--in-place --pre-summary-newline --wrap-descriptions 100 --wrap-summaries 100"
    )


@task()
def fmt(ctx: Context, target: t.Optional[str] = None, quiet: bool = False) -> None:
    """Autoformat code and docstrings."""
    if not quiet:
        print("Running docformatter")
    docformatter(ctx, target)

    if not quiet:
        print("Running isort")
    isort(ctx, target, quiet=True)

    if not quiet:
        print("Running black")
    black(ctx, target, quiet=True)


@task()
def flake8(ctx: Context) -> None:
    """Check code for PEP8 violations using flake8."""
    run(f"flake8 --format=pylint --exclude='{LINT_EXCLUDE}' {LINT_TARGETS}")


@task()
def pylint(ctx: Context) -> None:
    """Check code for static errors using pylint."""
    run(f"pylint --ignore='{LINT_EXCLUDE}' {LINT_TARGETS}")


@task()
def mypy(ctx: Context) -> None:
    """Check code using mypy type checker."""
    run(f"mypy --exclude='{LINT_EXCLUDE}' {LINT_TARGETS} --no-error-summary")


@task()
def lint(ctx: Context) -> None:
    """Run linters."""
    linters = {
        "flake8": flake8,
        "pylint": pylint,
        "mypy": mypy,
    }
    # in python 3.8 and before the ast module doesn't have the `unparse` function
    # which is needed for the generation
    if sys.version_info >= (3, 9):
        linters["chaining_types_update_required"] = chaining_types_update_required

    failures = []

    print(f"Preparing to run linters: {', '.join(linters)}\n")

    for name, linter in linters.items():
        print(f"Running {name}")
        try:
            linter(ctx)
        except EXIT_EXCEPTIONS:
            failures.append(name)
            result = "FAILED"
        else:
            result = "PASSED"
        print(f"{result}\n")

    if failures:
        failed = ", ".join(failures)
        raise Exit(f"ERROR: linters failed: {failed}")


@task(help={"args": "Override default pytest arguments"})
def test(
    ctx: Context, args: str = f"{TEST_TARGETS} --cov={PACKAGE_NAME}", with_mypy_tests: bool = False
) -> None:
    """Run unit tests using pytest."""
    tox_env_site_packages_dir = os.getenv("TOX_ENV_SITE_PACKAGES_DIR")
    if tox_env_site_packages_dir:
        # Re-path package source to match tox env so that we generate proper coverage report.
        tox_env_pkg_src = os.path.join(tox_env_site_packages_dir, os.path.basename(PACKAGE_SOURCE))
        args = args.replace(PACKAGE_SOURCE, tox_env_pkg_src)

    ignored_dirs = f"--ignore={MYPY_TESTS_DIR}" if with_mypy_tests is False else ""

    run(f"pytest {args} {ignored_dirs}")


@task()
def ci(ctx: Context) -> None:
    """Run linters and tests."""
    print("Building package")
    build(ctx)

    print("Building docs")
    docs(ctx)

    print("Checking linters")
    lint(ctx)

    print("Running unit tests")
    test(ctx)


@task()
def docs(ctx: Context, serve: bool = False, bind: str = "127.0.0.1", port: int = 8000) -> None:
    """Build docs."""
    run("rm -rf docs/_build")
    run("sphinx-build -q -W -b html docs docs/_build/html")

    if serve:
        print(f"Serving docs on {bind} port {port} (http://{bind}:{port}/) ...")
        run(f"python -m http.server -b {bind} --directory docs/_build/html {port}", hide=True)


@task()
def build(ctx: Context) -> None:
    """Build Python package."""
    run("rm -rf dist build docs/_build")
    run("python -m build")


@task()
def clean(ctx: Context) -> None:
    """Remove temporary files related to development."""
    run("find . -type f -name '*.py[cod]' -delete -o -type d -name __pycache__ -delete")
    run("rm -rf .tox .coverage .cache .pytest_cache **/.egg* **/*.egg* dist build .mypy_cache")


@task(pre=[build])
def release(ctx: Context) -> None:
    """Release Python package."""
    run("twine upload dist/*")


@task()
def generate_mypy_test(ctx: Context, file: str) -> None:
    """Generate base mypy test ready to be filled from doctests inside a python file."""
    run(
        "python scripts/mypy_doctests_generator.py"
        f" {file} tests/pytest_mypy_testing/test_{Path(file).name}"
    )


@task()
def generate_chaining_types(
    ctx: Context, output: str = "src/pydash/chaining/all_funcs.pyi"
) -> None:
    """Generates `all_funcs.pyi` stub file that types the chaining interface."""
    run(
        "python scripts/chaining_type_generator.py"
        f" --class_name AllFuncs --output {output} --wrapper Chain"
    )
    fmt(ctx, output, quiet=True)


@task()
def chaining_types_update_required(ctx: Context) -> None:
    with tempfile.NamedTemporaryFile(suffix=".pyi", dir=".") as tmp_file:
        generate_chaining_types(ctx, tmp_file.name)

        with open("src/pydash/chaining/all_funcs.pyi", "rb") as current_file:
            current = current_file.read()
        with open(tmp_file.name, "rb") as formatted_tmp_file:
            new = formatted_tmp_file.read()

        if current != new:
            err_msg = (
                "ERROR: src/pydash/chaining/all_funcs.pyi is out of date. Please run "
                "`inv generate-chaining-types` and commit the changes."
            )
            print(err_msg, file=sys.stderr)
            raise Exit()



import pytest

import pydash as _


parametrize = pytest.mark.parametrize


@parametrize(
    "case,expected",
    [
        ((5, 3), 8),
    ],
)
def test_add(case, expected):
    assert _.add(*case) == expected


@parametrize(
    "case,expected",
    [
        ((4.006,), 5),
        ((6.004, 2), 6.01),
        ((6040, -2), 6100),
        (([4.006, 6.004], 2), [4.01, 6.01]),
    ],
)
def test_ceil(case, expected):
    assert _.ceil(*case) == expected


@parametrize(
    "case,expected",
    [
        ((0, -1, 1), 0),
        ((1, -1, 1), 1),
        ((-1, -1, 1), -1),
        ((1, 1), 1),
        ((5, -1, 1), 1),
        ((-5, -1, 1), -1),
    ],
)
def test_clamp(case, expected):
    assert _.clamp(*case) == expected


@parametrize(
    "dividend,divisor,expected",
    [(10, 5, 2.0), (None, 1, 1.0), (None, None, 1.0), (1.5, 3, 0.5), (-10, 2, -5.0)],
)
def test_divide(dividend, divisor, expected):
    assert _.divide(dividend, divisor) == expected


@parametrize(
    "case,expected",
    [
        ((4.006,), 4),
        ((0.046, 2), 0.04),
        ((4060, -2), 4000),
        (([4.006, 0.046], 2), [4.0, 0.04]),
    ],
)
def test_floor(case, expected):
    assert _.floor(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3],), 3),
        (({"a": 3, "b": 2, "c": 1},), 3),
    ],
)
def test_max_(case, expected):
    assert _.max_(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3],), 3),
        (({"a": 3, "b": 2, "c": 1},), 3),
        ((["anaconda", "bison", "camel"], lambda x: len(x)), "anaconda"),
        (
            (
                [{"name": "barney", "age": 36}, {"name": "fred", "age": 40}],
                "age",
            ),
            {"name": "fred", "age": 40},
        ),
        (
            ([{"name": "barney", "age": 36}, {"name": "fred", "age": 40}], lambda c: c["age"]),
            {"name": "fred", "age": 40},
        ),
    ],
)
def test_max_by(case, expected):
    assert _.max_by(*case) == expected


@parametrize(
    "collection,default,expected",
    [([], -1, -1), ([1, 2, 3], -1, 3), ({}, -1, -1), ([], None, None), ({}, None, None)],
)
def test_max_default(collection, default, expected):
    assert _.max_(collection, default=default) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3, 4, 5], 3),
        ([0, 0.5, 1], 0.5),
    ],
)
def test_mean(case, expected):
    assert _.mean(case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3, 4, 5],), 3),
        (([{"b": 4}, {"b": 5}, {"b": 6}], "b"), 5),
        (([0, 0.5, 1],), 0.5),
        (({"one": {"a": 1}, "two": {"a": 2}, "three": {"a": 3}}, "a"), 2),
    ],
)
def test_mean_by(case, expected):
    assert _.mean_by(*case) == expected


@parametrize(
    "case,expected",
    [
        (([0, 0, 0, 0, 5],), 0),
        (([0, 0, 1, 2, 5],), 1),
        (([0, 0, 1, 2],), 0.5),
        (([0, 0, 1, 2, 3, 4],), 1.5),
    ],
)
def test_median(case, expected):
    assert _.median(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3],), 1),
        (({"a": 3, "b": 2, "c": 1},), 1),
    ],
)
def test_min_(case, expected):
    assert _.min_(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3],), 1),
        (({"a": 3, "b": 2, "c": 1},), 1),
        ((["anaconda", "bison", "cat"], lambda x: len(x)), "cat"),
        (
            (
                [{"name": "barney", "age": 36}, {"name": "fred", "age": 40}],
                "age",
            ),
            {"name": "barney", "age": 36},
        ),
        (
            ([{"name": "barney", "age": 36}, {"name": "fred", "age": 40}], lambda c: c["age"]),
            {"name": "barney", "age": 36},
        ),
    ],
)
def test_min_by(case, expected):
    assert _.min_by(*case) == expected


@parametrize(
    "collection,default,expected",
    [([], -1, -1), ([1, 2, 3], -1, 1), ({}, -1, -1), ([], None, None), ({}, None, None)],
)
def test_min_default(collection, default, expected):
    assert _.min_(collection, default=default) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3, 4, 5], 3), [2, 3, 4]),
        (([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3), [2, 3, 4, 5, 6, 7, 8, 9]),
        (([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 4), [2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]),
    ],
)
def test_moving_mean(case, expected):
    assert _.moving_mean(*case) == expected


@parametrize(
    "multiplier,multiplicand,expected",
    [(10, 5, 50), (None, 1, 1), (None, None, 1), (1.5, 3, 4.5), (-10, 2, -20)],
)
def test_multiply(multiplier, multiplicand, expected):
    assert _.multiply(multiplier, multiplicand) == expected


@parametrize(
    "case,expected",
    [
        ((2, 3), 8),
        ((3, 4), 81),
        (([1, 2, 3, 4, 5], 2), [1, 4, 9, 16, 25]),
        (("junk", 2), None),
    ],
)
def test_power(case, expected):
    assert _.power(*case) == expected


@parametrize(
    "case,expected",
    [
        ((2.51,), 3),
        ((2.499,), 2),
        ((2.499, 2), 2.50),
        (([2.5, 2.499, 2.555], 2), [2.50, 2.50, 2.56]),
        (("junk",), None),
    ],
)
def test_round_(case, expected):
    assert _.round_(*case) == expected


@parametrize(
    "case,expected",
    [
        (([2, 5, 10], 1), [0.2, 0.5, 1]),
        (([1, 2, 5], 1), [0.2, 0.4, 1]),
        (([1, 2, 5], 5), [1, 2, 5]),
        (([1, 2, 5],), [0.2, 0.4, 1]),
    ],
)
def test_scale(case, expected):
    assert _.scale(*case) == expected


@parametrize(
    "case,expected",
    [
        (([0, 0], [5, 5]), 1),
        (([0, 0], [1, 10]), 10),
        (([0, 0], [0, 10]), float("inf")),
        (([0, 0], [10, 0]), 0),
    ],
)
def test_slope(case, expected):
    assert _.slope(*case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], (2.0 / 3.0) ** 0.5),
    ],
)
def test_std_deviation(case, expected):
    assert _.std_deviation(case) == expected


@parametrize(
    "minuend,subtrahend,expected",
    [
        (10, 4, 6),
        (-6, -4, -2),
        (4, -10, 14),
        (-10, 4, -14),
        ("10", "5", 5),
        (2, 0.5, 1.5),
        (None, None, 0),
    ],
)
def test_subtract(minuend, subtrahend, expected):
    assert _.subtract(minuend, subtrahend) == expected


@parametrize(
    "minuend,subtrahend",
    [
        ("abs", 4),
        (4, "abc"),
        ("abs", "abc"),
        ([1, 2, 3, 4], 4),
    ],
)
def test_subtract_exception(minuend, subtrahend):
    with pytest.raises(TypeError):
        _.subtract(minuend, subtrahend)


@parametrize(
    "case,expected",
    [
        ([1, 2, 3, 4, 5], 15),
        ([0, 14, 0.2], 14.2),
    ],
)
def test_sum_(case, expected):
    assert _.sum_(case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3, 4, 5], lambda a: a * 2), 30),
        (([{"b": 4}, {"b": 5}, {"b": 6}], "b"), 15),
        (({"one": {"a": 1}, "two": {"a": 2}, "three": {"a": 3}}, "a"), 6),
    ],
)
def test_sum_by(case, expected):
    assert _.sum_by(*case) == expected


@parametrize(
    "case,expected",
    [
        ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[1, 4, 7], [2, 5, 8], [3, 6, 9]]),
    ],
)
def test_transpose(case, expected):
    assert _.transpose(case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], 2.0 / 3.0),
    ],
)
def test_variance(case, expected):
    assert _.variance(case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3],), [-1.225, 0.0, 1.225]),
        (([{"a": 1}, {"a": 2}, {"a": 3}], "a"), [-1.225, 0.0, 1.225]),
    ],
)
def test_zscore(case, expected):
    assert _.map_(_.zscore(*case), lambda v: round(v, 3)) == expected



import typing as t

import pytest

import pydash as _


class MyClass:
    def __init__(self) -> None:
        self.x = 5
        self.lst: t.List[int] = []

    def get_x(self) -> int:
        return self.x


@pytest.mark.mypy_testing
def test_mypy_assign() -> None:
    obj = {b'd': 5.5}
    reveal_type(_.assign(obj, {'a': 1}, {'b': 2}, {'c': 3}))  # R: builtins.dict[Union[builtins.bytes, builtins.str], Union[builtins.float, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_assign_with() -> None:
    def customizer(x: t.Union[int, None], y: int) -> float:
        return float(y) if x is None else float(x + y)

    reveal_type(_.assign_with({'a': 1}, {'b': 2}, {'a': 3}, customizer=customizer))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.float]]


@pytest.mark.mypy_testing
def test_mypy_callables() -> None:
    reveal_type(_.callables({'a': 1, 'b': lambda: 2, 'c': lambda: 3}))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_clone() -> None:
    reveal_type(_.clone({'hello': 5}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_clone_with() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {'a': 1, 'b': 2, 'c': {'d': 3}}

    def cbk(v: t.Union[int, t.Dict[str, int]], k: t.Union[str, None]) -> t.Union[int, None]:
        return v + 2 if isinstance(v, int) and k else None

    reveal_type(_.clone_with(x, cbk))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int], None]]


@pytest.mark.mypy_testing
def test_mypy_clone_deep() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {'a': 1, 'b': 2, 'c': {'d': 3}}
    reveal_type(_.clone_deep(x))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int]]]


@pytest.mark.mypy_testing
def test_mypy_clone_deep_with() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {'a': 1, 'b': 2, 'c': {'d': 3}}

    def cbk(v: t.Union[int, t.Dict[str, int]], k: t.Union[str, None]) -> t.Union[int, None]:
        return v + 2 if isinstance(v, int) and k else None

    reveal_type(_.clone_deep_with(x, cbk))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int], None]]


@pytest.mark.mypy_testing
def test_mypy_defaults() -> None:
    obj = {'a': 1}
    reveal_type(_.defaults(obj, {'b': 2}, {'c': 3}, {'a': 4}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_defaults_deep() -> None:
    obj = {'a': {'b': 1}}
    reveal_type(_.defaults_deep(obj, {'a': {'b': 2, 'c': 3}}))  # R: builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_find_key() -> None:
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.find_key({'a': 1, 'b': 2, 'c': 3}, is_one))  # R: Union[builtins.str, None]
    reveal_type(_.find_key([1, 2, 3, 4], is_one))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_find_last_key() -> None:
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.find_last_key({'a': 1, 'b': 2, 'c': 3}, is_one))  # R: Union[builtins.str, None]
    reveal_type(_.find_last_key([1, 2, 3, 1], is_one))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_for_in() -> None:
    def cb(v: int, k: str) -> None:
        return None
    reveal_type(_.for_in({'a': 1, 'b': 2, 'c': 3}, cb))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_for_in_right() -> None:
    def cb(v: int) -> None:
        return None

    reveal_type(_.for_in_right([1, 2, 3, 4], cb))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_get() -> None:
    reveal_type(_.get({}, 'a.b.c'))  # R: Any
    reveal_type(_.get({'a': {'b': {'c': [1, 2, 3, 4]}}}, 'a.b.c[1]'))  # R: Any
    reveal_type(_.get({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.2'))  # R: Any
    reveal_type(_.get(['a', 'b'], 0))  # R: Union[builtins.str, None]
    reveal_type(_.get(['a', 'b'], 0, 'c'))  # R: builtins.str
    reveal_type(_.get(MyClass(), 'x'))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_has() -> None:
    reveal_type(_.has([1, 2, 3], 1))  # R: builtins.bool
    reveal_type(_.has({'a': 1, 'b': 2}, 'b'))  # R: builtins.bool
    reveal_type(_.has(MyClass(), 'get_x'))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_invert() -> None:
    reveal_type(_.invert({'a': 1, 'b': 2, 'c': 3}))  # R: builtins.dict[builtins.int, builtins.str]


@pytest.mark.mypy_testing
def test_mypy_invert_by() -> None:
    obj = {'a': 1, 'b': 2, 'c': 1}
    reveal_type(_.invert_by(obj))  # R: builtins.dict[builtins.int, builtins.list[builtins.str]]

    def group_prefix(x: int) -> str:
        return 'group' + str(x)

    reveal_type(_.invert_by(obj, group_prefix))  # R: builtins.dict[builtins.str, builtins.list[builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_invoke() -> None:
    obj = {'a': [{'b': {'c': [1, 2, 3, 4]}}]}
    reveal_type(_.invoke(obj, 'a[0].b.c.pop', 1))  # R: Any
    reveal_type(_.invoke(MyClass(), 'get_x'))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_keys() -> None:
    reveal_type(_.keys([1, 2, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.keys({'a': 1, 'b': 2}))  # R: builtins.list[builtins.str]
    reveal_type(_.keys(MyClass()))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_map_keys() -> None:
    def callback(value: int, key: str) -> str:
        return key * 2

    reveal_type(_.map_keys({'a': 1, 'b': 2, 'c': 3}, callback))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_map_values() -> None:
    def times_two(x: int) -> str:
        return str(x * 2)

    reveal_type(_.map_values({'a': 1, 'b': 2, 'c': 3}, times_two))  # R: builtins.dict[builtins.str, builtins.str]
    reveal_type(_.map_values({'a': 1, 'b': {'d': 4}, 'c': 3}, {'d': 4}))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_map_values_deep() -> None:
    x = {'a': 1, 'b': {'c': 2}}

    def times_two(x: int) -> int:
        return x * 2

    reveal_type(_.map_values_deep(x, times_two))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_merge() -> None:
    obj = {'a': 2}
    reveal_type(_.merge(obj, {'a': 1}, {'b': 2, 'c': 3}, {'d': 4}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_merge_with() -> None:
    cbk = lambda obj_val, src_val: obj_val + src_val
    obj1 = {'a': [1], 'b': [2]}
    obj2 = {'a': [3], 'b': [4]}
    reveal_type(_.merge_with(obj1, obj2, cbk))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_omit() -> None:
    reveal_type(_.omit({'a': 1, 'b': 2, 'c': 3}, 'b', 'c'))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.omit({'a': 1, 'b': 2, 'c': 3 }, ['a', 'c']))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.omit([1, 2, 3, 4], 0, 3))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.omit({'a': {'b': {'c': 'd'}}}, 'a.b.c'))  # R: builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.str]]]
    reveal_type(_.omit(MyClass(), 'x'))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_omit_by() -> None:
    def is_int(v: t.Union[str, int]) -> bool:
        return isinstance(v, int)

    obj: t.Dict[str, t.Union[str, int]] = {'a': 1, 'b': '2', 'c': 3}
    reveal_type(_.omit_by(obj, is_int))  # R: builtins.dict[builtins.str, Union[builtins.str, builtins.int]]
    reveal_type(_.omit_by([1, 2, 3, 4], is_int))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.omit_by(MyClass(), is_int))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_parse_int() -> None:
    reveal_type(_.parse_int('5'))  # R: Union[builtins.int, None]
    reveal_type(_.parse_int('12', 8))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_pick() -> None:
    reveal_type(_.pick({'a': 1, 'b': 2, 'c': 3}, 'a', 'b'))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.pick(MyClass(), 'x'))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_pick_by() -> None:
    def is_int(v: t.Union[int, str]) -> bool:
        return isinstance(v, int)

    obj: t.Dict[str, t.Union[int, str]] = {'a': 1, 'b': '2', 'c': 3 }

    reveal_type(_.pick_by(obj, is_int))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.str]]
    reveal_type(_.pick(MyClass(), lambda v: isinstance(v, int)))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_rename_keys() -> None:
    reveal_type(_.rename_keys({'a': 1, 'b': 2, 'c': 3}, {'a': 'A', 'b': 'B'}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_set_() -> None:
    reveal_type(_.set_({}, 'a.b.c', 1))  # R: builtins.dict[<nothing>, <nothing>]
    reveal_type(_.set_(MyClass(), "x", 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_set_with() -> None:
    reveal_type(_.set_with({}, '[0][1]', 'a', lambda: {}))  # R: builtins.dict[<nothing>, <nothing>]
    reveal_type(_.set_with(MyClass(), "x", lambda: 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_to_boolean() -> None:
    reveal_type(_.to_boolean('true'))  # R: Union[builtins.bool, None]


@pytest.mark.mypy_testing
def test_mypy_to_dict() -> None:
    obj = {'a': 1, 'b': 2}
    reveal_type(_.to_dict(obj))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.to_dict(MyClass()))  # R: builtins.dict[Any, Any]
    reveal_type(_.to_dict([1, 2, 3, 4]))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.to_dict([(1, 2), (3, 4)]))  # R: builtins.dict[builtins.int, Tuple[builtins.int, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_to_integer() -> None:
    reveal_type(_.to_integer(3.2))  # R: builtins.int
    reveal_type(_.to_integer('3.2'))  # R: builtins.int
    reveal_type(_.to_integer('invalid'))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_to_list() -> None:
    reveal_type(_.to_list({'a': 1, 'b': 2, 'c': 3}))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list((1, 2, 3, 4)))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list(1))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list([1]))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list(a for a in [1, 2, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list('cat'))  # R: builtins.list[builtins.str]
    reveal_type(_.to_list('cat', split_strings=False))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_to_number() -> None:
    reveal_type(_.to_number('1234.5678'))  # R: Union[builtins.float, None]
    reveal_type(_.to_number('1234.5678', 4))  # R: Union[builtins.float, None]
    reveal_type(_.to_number(1, 2))  # R: Union[builtins.float, None]


@pytest.mark.mypy_testing
def test_mypy_to_pairs() -> None:
    reveal_type(_.to_pairs([1, 2, 3, 4]))  # R: builtins.list[builtins.list[builtins.int]]
    reveal_type(_.to_pairs({'a': 1}))  # R: builtins.list[builtins.list[Union[builtins.str, builtins.int]]]
    reveal_type(_.to_pairs(MyClass()))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_to_string() -> None:
    reveal_type(_.to_string(1))  # R: builtins.str
    reveal_type(_.to_string(None))  # R: builtins.str
    reveal_type(_.to_string([1, 2, 3]))  # R: builtins.str
    reveal_type(_.to_string('a'))  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_transform() -> None:
    def build_list(acc: t.List[t.Tuple[int, int]], v: int, k: int) -> None:
        return acc.append((k, v))

    base_list: t.List[t.Tuple[int, int]] = []
    reveal_type(_.transform([1, 2, 3, 4], build_list, base_list))  # R: builtins.list[Tuple[builtins.int, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_update() -> None:
    reveal_type(_.update({}, ['a', 'b'], lambda value: value))  # R: builtins.dict[Any, Any]
    reveal_type(_.update([], [0, 0], lambda value: 1))  # R: builtins.list[Any]
    reveal_type(_.update(MyClass(), 'x', lambda value: 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_update_with() -> None:
    reveal_type(_.update_with({}, '[0][1]', lambda x: 'a', lambda x: {}))  # R: builtins.dict[Any, Any]
    reveal_type(_.update_with([], [0, 0], lambda x: 1, lambda x: []))  # R: builtins.list[Any]
    reveal_type(_.update_with(MyClass(), 'lst.0', lambda value: 10, lambda x: []))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_unset() -> None:
    reveal_type(_.unset({'a': [{'b': {'c': 7}}]}, 'a[0].b.c'))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_values() -> None:
    reveal_type(_.values({'a': 'a', 'b': 'b', 'c': 'c'}))  # R: builtins.list[builtins.str]
    reveal_type(_.values({'a': 1, 'b': 2, 'c': 3}))  # R: builtins.list[builtins.int]
    reveal_type(_.values([2, 4, 6, 8]))  # R: builtins.list[builtins.int]
    reveal_type(_.values(MyClass()))  # R: builtins.list[Any]
